/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/
/*
  Alle Bloecke, die im CodedImage durch 'm_CodeXX[X]' gekennzeichnet sind,
  werden transformiert. Danach koennen diese Flags dann immer noch auf
  false gesetzt werden.
*/

#ifndef CODING_DCT_HH
#define CODING_DCT_HH

#include "image/image.hh"
#include "control/datapool.hh"
#include "control/taskpool.hh"

template <class Pel> class DCTCoder
{
public:
  DCTCoder() { }
  virtual ~DCTCoder() { }
  
  virtual void SetChromaFormat(uint2 chroma_format) = 0;
  // virtual void SetQuantizerMatrix(enum which,Matrix&) = 0;

  /* Fuehrt auf allen Bloecken der Eingabebildes 'img' eine FDCT durch und Quantisiert die Ausgabe.
     Die Koeffizienten werden in das Ausgabebild 'cimg' geschrieben. */
  virtual void CodeDCTBlocks(const Image<Pel>& img,ImageData& cimg);

  /* Folgende beiden Methoden dienen der Codierung einzelner Makrobloecke. Vor der eigentlichen
     Codierung muss 'SetQScaleType()' und 'SetDCPrecision()' aufgerufen werden.
     Danach koennen mit 'CodeDCTBlock' einzelne Makrobloecke aus einem Bild codiert werden.
     Der Makroblock wird (in Makroblockkoordinaten) ueber 'mbx0','mby0'
     ausgewaehlt und die Daten werden in 'mb' gespeichert. */
  virtual void SetQScaleType (uint1 qscale_type) = 0;
  virtual void SetDCPrecision(uint8 dc_precision) = 0;

  /* Codieren eines MBs.
  */
  virtual void MBlk_FDCTQuant(const Image<Pel>& img,int mbx0,int mby0,Macroblock& mb) = 0;
  virtual void MBlk_FDCT (const Image<Pel>&,int mbx0,int mby0,Macroblock& mb) = 0;
  virtual void MBlk_Quant(const Macroblock& in,Macroblock& out) = 0;
};


class Task_DCTQuantize : public Task
{
public:
  Task_DCTQuantize(int framenr);

  void Execute(ProcessData&);

  const int m_framenr;
};

#endif
