/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "types.hh"
#include "coding/dct_separated.hh"

#include "coding/utility.hh"
#include "mpeg/defaultmatrices.hh"
#include "mpeg/constants.hh"
#include "options.hh"

#include <iostream.h>
#include <stdlib.h>


#define easysign(x) ((x>=0) ? 1 : -1)


template <class Pel> void DCTCoder_Separated<Pel>::SetChromaFormat(uint2 chroma_format)
{
  switch (chroma_format)
    {
    case CHROMA_420: d_chromaxfact= 8; d_chromayfact= 8; break;
    case CHROMA_422: d_chromaxfact= 8; d_chromayfact=16; break;
    case CHROMA_444: d_chromaxfact=16; d_chromayfact=16; break;
    }
}

template <class Pel> void DCTCoder_Separated<Pel>::SetQScaleType(uint1 qscale_type)
{
  d_recalc_scalefacts |= (qscale_type != d_q_scale_type);
  d_q_scale_type       = qscale_type;
}

template <class Pel> void DCTCoder_Separated<Pel>::SetDCPrecision(uint8 dc_precision)
{
  d_recalc_scalefacts |= (dc_precision != d_dc_precision);
  d_dc_precision       = dc_precision;
}


template <class Pel> void DCTCoder_Separated<Pel>::MBlk_FDCT(const Image<Pel>& srcimg,
                                                             int x,int y,Macroblock& mb)
{
  /*  BILD CODIEREN */

  const Pel*const * y_pix  = srcimg.AskBitmap_const(Bitmap_Y ).AskFrame_const();
  const Pel*const * cr_pix = srcimg.AskBitmap_const(Bitmap_Cr).AskFrame_const();
  const Pel*const * cb_pix = srcimg.AskBitmap_const(Bitmap_Cb).AskFrame_const();


  const Pel* in[8];

  int lx0 = x*16,  lx1 = x*16+8;
  int ly0 = y*16,  ly1 = y*16+8;

  int cx0 = x*d_chromaxfact , cx1 = x*d_chromaxfact+8;
  int cy0 = y*d_chromayfact , cy1 = y*d_chromayfact+8;

  if (mb.m_CodeY[0])
    {
      {for (int i=0;i<8;i++) { in[i] = &y_pix[ly0+i][lx0]; }}
      d_dct.PerformFDCT2(in,mb.m_y[0]);
    }

  if (mb.m_CodeY[1])
    {
      {for (int i=0;i<8;i++) { in[i] = &y_pix[ly0+i][lx1]; }}
      d_dct.PerformFDCT2(in,mb.m_y[1]);
    }

  if (mb.m_CodeY[2])
    {
      {for (int i=0;i<8;i++) { in[i] = &y_pix[ly1+i][lx0]; }}
      d_dct.PerformFDCT2(in,mb.m_y[2]);
    }

  if (mb.m_CodeY[3])
    {
      {for (int i=0;i<8;i++) { in[i] = &y_pix[ly1+i][lx1]; }}
      d_dct.PerformFDCT2(in,mb.m_y[3]);
    }

  if (mb.m_CodeCr[0])
    {
      {for (int i=0;i<8;i++) { in[i] = &cr_pix[cy0+i][cx0]; }}
      d_dct.PerformFDCT2(in,mb.m_cr[0]);
    }

  if (mb.m_CodeCr[1])
    {
      {for (int i=0;i<8;i++) { in[i] = &cr_pix[cy1+i][cx0]; }}
      d_dct.PerformFDCT2(in,mb.m_cr[1]);
    }

  if (mb.m_CodeCr[2])
    {
      {for (int i=0;i<8;i++) { in[i] = &cr_pix[cy0+i][cx1]; }}
      d_dct.PerformFDCT2(in,mb.m_cr[2]);
    }

  if (mb.m_CodeCr[3])
    {
      {for (int i=0;i<8;i++) { in[i] = &cr_pix[cy1+i][cx1]; }}
      d_dct.PerformFDCT2(in,mb.m_cr[3]);
    }

  if (mb.m_CodeCb[0])
    {
      {for (int i=0;i<8;i++) { in[i] = &cb_pix[cy0+i][cx0]; }}
      d_dct.PerformFDCT2(in,mb.m_cb[0]);
    }

  if (mb.m_CodeCb[1])
    {
      {for (int i=0;i<8;i++) { in[i] = &cb_pix[cy1+i][cx0]; }}
      d_dct.PerformFDCT2(in,mb.m_cb[1]);
    }

  if (mb.m_CodeCb[2])
    {
      {for (int i=0;i<8;i++) { in[i] = &cb_pix[cy0+i][cx1]; }}
      d_dct.PerformFDCT2(in,mb.m_cb[2]);
    }

  if (mb.m_CodeCb[3])
    {
      {for (int i=0;i<8;i++) { in[i] = &cb_pix[cy1+i][cx1]; }}
      d_dct.PerformFDCT2(in,mb.m_cb[3]);
    }
}


template <class Pel> void DCTCoder_Separated<Pel>::MBlk_Quant(const Macroblock& mb,Macroblock& out)
{
  if (d_recalc_scalefacts)
    {
      PrecalcQMatrix();
    }

  // int qscale = qcode2qscale[d_q_scale_type] [mb.m_quantiser_scale_code];
 
  if (out.m_IsIntra)
    {
      int dc_offset = (1<<(11-1-d_dc_precision));
          
      for (int blk=0;blk<12;blk++)
        if (mb.m_CodeBlk[blk])
          {
            const DCTBlock& dctblk_in  = mb.m_blks[blk];
                  DCTBlock& dctblk_out = out.m_blks[blk];
            
            for (int u=0;u<8;u++)
              for (int v=0;v<8;v++)
                {
                  if (u==0 && v==0)
                    {
                      if (dctblk_in.m_coeff[0][0] >=0)
                        dctblk_out.m_coeff[0][0] = dctblk_in.m_coeff[0][0] + dc_offset;
                      else
                        dctblk_out.m_coeff[0][0] = dctblk_in.m_coeff[0][0] - dc_offset;

                      dctblk_out.m_coeff[0][0] = dctblk_in.m_coeff[0][0] >> (11-d_dc_precision);
                    }
                  else
                    {
                      dctblk_out.m_coeff[u][v] =
                        (32*(int)dctblk_in.m_coeff[u][v] +
                         easysign(dctblk_in.m_coeff[u][v])*d_QMatrix[out.m_quantiser_scale_code][u][v])/
                        (2*d_QMatrix[out.m_quantiser_scale_code][u][v]);
                    }
                }
          }
        }
  else
    {
      for (int blk=0;blk<12;blk++)
        if (mb.m_CodeBlk[blk])
          {
            const DCTBlock& dctblk_in  = mb.m_blks[blk];
                  DCTBlock& dctblk_out = out.m_blks[blk];

            for (int u=0;u<8;u++)
              for (int v=0;v<8;v++)
                {
                  dctblk_out.m_coeff[u][v] = d_inter[mb.m_quantiser_scale_code][dctblk_in.m_coeff[u][v]];
                }
          }
    }


#if 0             // ------------------ Hochfrequenze Koeffizienten auf 0 setzen. -------------
  for (int blk=0;blk<12;blk++)
    if (mb.m_CodeBlk[blk])
      {
        DCTBlock& dctblk_out = out.m_blks[blk];
            
        for (int u=0;u<8;u++)
          for (int v=0;v<8;v++)
            if (u+v>5)
              {
                dctblk_out.m_coeff[u][v] = 0;
              }
      }
#endif

#if 0     // --------------------- Block streichen, wenn Koeffizientensumme unter Threshold liegt
  for (int blk=0;blk<12;blk++)
    if (mb.m_CodeBlk[blk])
      {
        DCTBlock& dctblk_out = out.m_blks[blk];

        int sum=0;
        for (int u=0;u<8;u++)
          for (int v=0;v<8;v++)
            {
              sum += abs(dctblk_out.m_coeff[u][v]);
            }

        if (sum<=5)
          {
            for (int u=0;u<8;u++)
              for (int v=0;v<8;v++)
                dctblk_out.m_coeff[u][v] = 0;
          }
      }
#endif

  // Der Wertebereich bei MPEG1 ist viel staerker eingeschraenkt !

  int bound;

  if (options.CreateMPEG2) bound=2047;
  else                     bound=255;

  for (int blk=0;blk<12;blk++)
    if (mb.m_CodeBlk[blk])
      {
        DCTBlock& dctblk_out = out.m_blks[blk];
        
        for (int u=0;u<8;u++)
          for (int v=0;v<8;v++)
            {
              if (dctblk_out.m_coeff[u][v]<-bound)
                {
                  cerr << "WARNING: DCT coefficient out of allowed range (< -" << bound << ")\n";
                  dctblk_out.m_coeff[u][v]=-bound;
                }
              else if (dctblk_out.m_coeff[u][v]>bound)
                {
                  cerr << "WARNING: DCT coefficient out of allowed range (> " << bound << ")\n";
                  dctblk_out.m_coeff[u][v]=bound;
                }
            }
      }
}


template <class Pel> void DCTCoder_Separated<Pel>::PrecalcQMatrix()
{
  {
    for (int qcode=1;qcode<32;qcode++)
      {
        int qscale = qcode2qscale[d_q_scale_type] [qcode];
        
        for (int v=0;v<8;v++)
          for (int u=0;u<8;u++)
            d_QMatrix[qcode][v][u] = qscale*QuantMatrixIntra[v][u];
      }
  }

  {
    for (int qcode=1;qcode<32;qcode++)
      {
        d_inter[qcode] = &d_qinter[qcode][2048];
        int qscale = qcode2qscale[d_q_scale_type] [qcode];
        
        for (int i=-2048;i<2048;i++)
          d_inter[qcode][i] = i/qscale; // *16/16
      }
  }

  d_recalc_scalefacts = false;
}

template class DCTCoder_Separated<Pixel>;
template class DCTCoder_Separated<DiffPixel>;
