/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "coding/decoder.hh"
#include "coding/utility.hh"
#include "mpeg/defaultmatrices.hh"
#include "mpeg/constants.hh"
#include "options.hh"

#include "fstream.h"
#include "visualize/writeppm.hh"

#include <iostream.h>
#include <iomanip.h>

inline void  Swap(Image<Pixel>*& a,Image<Pixel>*& b) { Image<Pixel>* c=a; a=b; b=c; }

IDCT2_AraiAguiNakajima_Int Task_Decode::decoder_dct;


void Task_Decode::Execute(ProcessData& procdata)
{
  ImageData* img = ImageData::GetNr(m_framenr);
  assert(img);

  if (options.ProgressDisplay >= Options::ProgressDisplay_Normal)
    {
      ShowTaskInformation("decode",procdata.threadnr);
      cerr << "frame " << m_framenr << endl;
      EndShowTaskInformation();
    }

  int mbRows    = img->mblks.AskHeight();
  int mbColumns = img->mblks.AskWidth();
  int width  = mbColumns*16;
  int height = mbRows*16;

  img->decoded.CreateImage(YUV,width,height);

  Image<Pixel>* previmg = NULL; if (img->prev_ip) { previmg=img->prev_ip->source; assert(previmg); }
  Image<Pixel>* nextimg = NULL; if (img->next_ip) { nextimg=img->next_ip->source; assert(nextimg); }

  Pixel*const * y_pix  = img->decoded.AskBitmap(Bitmap_Y ).AskFrame();
  Pixel*const * cr_pix = img->decoded.AskBitmap(Bitmap_Cr).AskFrame();
  Pixel*const * cb_pix = img->decoded.AskBitmap(Bitmap_Cb).AskFrame();
  //cout << "AAA\n";
  for (int mby=0;mby<mbRows;mby++)
    for (int mbx=0;mbx<mbColumns;mbx++)
      {
	const Macroblock& mb = img->mblks.Ask_const(mby,mbx);

	int lx0 = mbx*16,  lx1 = mbx*16+8;
	int ly0 = mby*16,  ly1 = mby*16+8;

        int chromaxfact=8; // TODO
        int chromayfact=8;
        int chromaxdiv =2;
        int chromaydiv =2;

	int cx0 = mbx*chromaxfact , cx1 = mbx*chromaxfact+8;
	int cy0 = mby*chromayfact , cy1 = mby*chromayfact+8;

	Pixel* out[8];

	bool  add_idct=true;


        //  cout << "C2\n";
	if (mb.m_HasMotionForward && mb.m_HasMotionBackward)
	  {
	    MotionCompensateDoubleSet(y_pix,lx0,ly0,lx0,ly0,16,16,
				      previmg->AskBitmap_const(Bitmap_Y).AskFrame_const(),
				      mb.m_forward1.m_habs,mb.m_forward1.m_vabs,
				      nextimg->AskBitmap_const(Bitmap_Y).AskFrame_const(),
				      mb.m_backward1.m_habs,mb.m_backward1.m_vabs);
	    MotionCompensateDoubleSet(cr_pix,cx0,cy0,cx0,cy0,chromaxfact,chromayfact,
				      previmg->AskBitmap_const(Bitmap_Cr).AskFrame_const(),
				      mb.m_forward1.m_habs/chromaxdiv,
				      mb.m_forward1.m_vabs/chromaydiv,
				      nextimg->AskBitmap_const(Bitmap_Cr).AskFrame_const(),
				      mb.m_backward1.m_habs/chromaxdiv,
				      mb.m_backward1.m_vabs/chromaydiv);
	    MotionCompensateDoubleSet(cb_pix,cx0,cy0,cx0,cy0,chromaxfact,chromayfact,
				      previmg->AskBitmap_const(Bitmap_Cb).AskFrame_const(),
				      mb.m_forward1.m_habs/chromaxdiv,
				      mb.m_forward1.m_vabs/chromaydiv,
				      nextimg->AskBitmap_const(Bitmap_Cb).AskFrame_const(),
				      mb.m_backward1.m_habs/chromaxdiv,
				      mb.m_backward1.m_vabs/chromaydiv);
	  }
	else if (mb.m_HasMotionForward)
	  {
	    MotionCompensateSingleSet(y_pix,lx0,ly0,lx0,ly0,16,16,
				      previmg->AskBitmap_const(Bitmap_Y).AskFrame_const(),
				      mb.m_forward1.m_habs,mb.m_forward1.m_vabs);
	    MotionCompensateSingleSet(cr_pix,cx0,cy0,cx0,cy0,chromaxfact,chromayfact,
				      previmg->AskBitmap_const(Bitmap_Cr).AskFrame_const(),
				      mb.m_forward1.m_habs/chromaxdiv,
				      mb.m_forward1.m_vabs/chromaydiv);
	    MotionCompensateSingleSet(cb_pix,cx0,cy0,cx0,cy0,chromaxfact,chromayfact,
				      previmg->AskBitmap_const(Bitmap_Cb).AskFrame_const(),
				      mb.m_forward1.m_habs/chromaxdiv,
				      mb.m_forward1.m_vabs/chromaydiv);
	  }
	else if (mb.m_HasMotionBackward)
	  {
	    MotionCompensateSingleSet(y_pix,lx0,ly0,lx0,ly0,16,16,
				      nextimg->AskBitmap_const(Bitmap_Y).AskFrame_const(),
				      mb.m_backward1.m_habs,mb.m_backward1.m_vabs);
	    MotionCompensateSingleSet(cr_pix,cx0,cy0,cx0,cy0,chromaxfact,chromayfact,
				      nextimg->AskBitmap_const(Bitmap_Cr).AskFrame_const(),
				      mb.m_backward1.m_habs/chromaxdiv,
				      mb.m_backward1.m_vabs/chromaydiv);
	    MotionCompensateSingleSet(cb_pix,cx0,cy0,cx0,cy0,chromaxfact,chromayfact,
				      nextimg->AskBitmap_const(Bitmap_Cb).AskFrame_const(),
				      mb.m_backward1.m_habs/chromaxdiv,
				      mb.m_backward1.m_vabs/chromaydiv);
	  }
	else
	  {
	    add_idct = false;
	  }

        //  cout << "C1\n";
	// IDCT ausfuehren

	const uint1 inter = (mb.m_IsIntra ? 0 : 1);
	const  int  q = mb.m_quantiser_scale_code;

	{for (int i=0;i<8;i++) { out[i] = &y_pix[ly0+i][lx0]; }}
	if (mb.m_CodeY[0]) IDCT(mb.m_y[0],out,inter,0,q,add_idct,img->pichdr);
	{for (int i=0;i<8;i++) { out[i] = &y_pix[ly0+i][lx1]; }}
	if (mb.m_CodeY[1]) IDCT(mb.m_y[1],out,inter,0,q,add_idct,img->pichdr);
	{for (int i=0;i<8;i++) { out[i] = &y_pix[ly1+i][lx0]; }}
	if (mb.m_CodeY[2]) IDCT(mb.m_y[2],out,inter,0,q,add_idct,img->pichdr);
	{for (int i=0;i<8;i++) { out[i] = &y_pix[ly1+i][lx1]; }}
	if (mb.m_CodeY[3]) IDCT(mb.m_y[3],out,inter,0,q,add_idct,img->pichdr);

	{for (int i=0;i<8;i++) { out[i] = &cr_pix[cy0+i][cx0]; }}
	if (mb.m_CodeCr[0]) IDCT(mb.m_cr[0],out,inter,1,q,add_idct,img->pichdr);
	{for (int i=0;i<8;i++) { out[i] = &cr_pix[cy1+i][cx0]; }}
	if (mb.m_CodeCr[1]) IDCT(mb.m_cr[1],out,inter,1,q,add_idct,img->pichdr);
	{for (int i=0;i<8;i++) { out[i] = &cr_pix[cy0+i][cx1]; }}
	if (mb.m_CodeCr[2]) IDCT(mb.m_cr[2],out,inter,1,q,add_idct,img->pichdr);
	{for (int i=0;i<8;i++) { out[i] = &cr_pix[cy1+i][cx1]; }}
	if (mb.m_CodeCr[3]) IDCT(mb.m_cr[3],out,inter,1,q,add_idct,img->pichdr);

	{for (int i=0;i<8;i++) { out[i] = &cb_pix[cy0+i][cx0]; }}
	if (mb.m_CodeCb[0]) IDCT(mb.m_cb[0],out,inter,1,q,add_idct,img->pichdr);
	{for (int i=0;i<8;i++) { out[i] = &cb_pix[cy1+i][cx0]; }}
	if (mb.m_CodeCb[1]) IDCT(mb.m_cb[1],out,inter,1,q,add_idct,img->pichdr);
	{for (int i=0;i<8;i++) { out[i] = &cb_pix[cy0+i][cx1]; }}
	if (mb.m_CodeCb[2]) IDCT(mb.m_cb[2],out,inter,1,q,add_idct,img->pichdr);
	{for (int i=0;i<8;i++) { out[i] = &cb_pix[cy1+i][cx1]; }}
	if (mb.m_CodeCb[3]) IDCT(mb.m_cb[3],out,inter,1,q,add_idct,img->pichdr);
      }

  //cout << "CCC\n";
  for (int y=0;y<height;y++)
    for (int x=0;x<width;x++)
      {
	// TODO: Im Farbbereich muss je nach ChomaFormat weniger Saturated werden.
	short pix;
	pix =  y_pix[y][x]; if (pix<0)  y_pix[y][x]=0; else if (pix>255)  y_pix[y][x]=255;
	pix = cr_pix[y][x]; if (pix<0) cr_pix[y][x]=0; else if (pix>255) cr_pix[y][x]=255;
	pix = cb_pix[y][x]; if (pix<0) cb_pix[y][x]=0; else if (pix>255) cb_pix[y][x]=255;
      }

#if 0
static pn=0;
	{
	  char buf[100];
	  sprintf(buf,"/msdos/vis/out%03d.ppm",pn); pn++;
	  ofstream ostr(buf);
	  Image<Pixel>* vimg = CreateVisualizationImage(img->decoded);
	  WriteImageAsPPM(*vimg,ostr);
	  delete vimg;
	}
#endif

        //cout << "BBB\n";
}

void Task_Decode::IDCT(const DCTBlock& blk,uint8* out[8], uint1 inter,uint1 chrom,int qcode,bool add,
                       PictureHeaderData& pichdr)
{
  DCTBlock qblk;

#if 0
  cout << endl;
  for (int v=0;v<8;v++,cout << endl)
    for (int u=0;u<8;u++)
      cout << blk.m_coeff[v][u] << ' ';
  cout << endl;
#endif

#if 0         // URSPRUENGLICHE VERSION
  if (!inter)
    {
      {for (int v=0;v<8;v++)
	for (int u=0;u<8;u++)
	  qblk.m_coeff[v][u] = blk.m_coeff[v][u] * qW[inter][chrom][qcode][v][u] / 16;}

      qblk.m_coeff[0][0] = blk.m_coeff[0][0] * qW[inter][chrom][qcode][0][0];
    }
  else
    {
      {for (int v=0;v<8;v++)
	for (int u=0;u<8;u++)
	  qblk.m_coeff[v][u] = (blk.m_coeff[v][u]*2 + Sign(blk.m_coeff[v][u]))
                               * qW[inter][chrom][qcode][v][u] / 32;}
    }
#endif

  // Inverse Quantisierung

  if (!inter)
    {
      {for (int v=0;v<8;v++)
	for (int u=0;u<8;u++)
	  qblk.m_coeff[v][u] = blk.m_coeff[v][u] * QuantMatrixIntra[v][u] *
            qcode2qscale[pichdr.m_q_scale_type][qcode] / 16;}

      qblk.m_coeff[0][0] = blk.m_coeff[0][0] * (1<<(11-pichdr.m_intra_dc_precision));
    }
  else
    {
      {for (int v=0;v<8;v++)
	for (int u=0;u<8;u++)
	  qblk.m_coeff[v][u] = (blk.m_coeff[v][u]*2 + Sign(blk.m_coeff[v][u])) *
            16 /* <- matrix*/ * qcode2qscale[pichdr.m_q_scale_type][qcode] / 32;}
    }


  // Saturation

  int sum=0;
  for (int v=0;v<8;v++)
    for (int u=0;u<8;u++)
      {
             if (qblk.m_coeff[v][u]> 2047) qblk.m_coeff[v][u]= 2047;
	else if (qblk.m_coeff[v][u]<-2048) qblk.m_coeff[v][u]=-2048;

	sum += qblk.m_coeff[v][u];
      }

  if ((sum&1)==0) { qblk.m_coeff[7][7] ^= 1; }
#if 0
  for (int v=0;v<8;v++,cout << endl)
    for (int u=0;u<8;u++)
      cout << qblk.m_coeff[v][u] << ' ';
  cout << endl;
#endif

#if 0
  for (int i=0;i<8;i++)
    cout << "out[" << i << "] : " << (void*)out[i] << endl;
#endif

  decoder_dct.PerformIDCT2Blk(qblk,out,add);

#if 0
  for (int v=0;v<8;v++,cout << endl)
    for (int u=0;u<8;u++)
      cout << 'a' << ((int)out[v][u]) << ' ';
  cout << endl;
#endif
}


void MotionCompensateSingleSet(Pixel*const* destination,
			       int x1,int y1,
			       int x0,int y0,int w,int h,
			       const Pixel*const* reference, int hdist,int vdist)
{
  bool half_h = ((hdist&1)!=0);
  bool half_v = ((vdist&1)!=0);

  int inthdist = hdist/2;
  int intvdist = vdist/2;

  int xsign = Sign(hdist);
  int ysign = Sign(vdist);

  int yend = y1+h;

  intvdist = y0-y1+intvdist;
  inthdist = x0-x1+inthdist;

  if (!half_h && !half_v)
    {
      for (int y=y1;y<yend;y++)
	{
	  const Pixel* rp = &reference[y+intvdist][inthdist+x1];
	        Pixel* dp = &destination[y][x1];

	  for (int i=w;i>0;i--)
	    *dp++ = *rp++;
	}
      return;
    }
  
  if (half_h && !half_v)
    {
      for (int y=y1;y<yend;y++)
	{
	  const Pixel* rp1 = &reference[y+intvdist][inthdist+x1];
	  const Pixel* rp2 = &reference[y+intvdist][inthdist+x1+xsign];
	        Pixel* dp  = &destination[y][x1];

	  for (int i=w;i>0;i--)
	    *dp++ = (*rp1++ + *rp2++ +1)>>1;
	}
      return;
    }

  if (!half_h && half_v)
    {
      for (int y=y1;y<yend;y++)
	{
	  const Pixel* rp1 = &reference[y+intvdist      ][inthdist+x1];
	  const Pixel* rp2 = &reference[y+intvdist+ysign][inthdist+x1];
	        Pixel* dp  = &destination[y][x1];

	  for (int i=w;i>0;i--)
	    *dp++ = (((short)*rp1++) + *rp2++ +1)>>1;
	}
      return;
    }

  if (half_h && half_v)
    {
      for (int y=y1;y<yend;y++)
	{
	  const Pixel* rp1 = &reference[y+intvdist      ][inthdist+x1];
	  const Pixel* rp2 = &reference[y+intvdist+ysign][inthdist+x1];
	  const Pixel* rp3 = &reference[y+intvdist      ][inthdist+x1+xsign];
	  const Pixel* rp4 = &reference[y+intvdist+ysign][inthdist+x1+xsign];
	        Pixel* dp  = &destination[y][x1];

	  for (int i=w;i>0;i--)
	    *dp++ = (((short)*rp1++) + *rp2++ + *rp3++ + *rp4++ + 2)>>2;
	}
      return;
    }
}


void MotionCompensateDoubleSet(Pixel*const* destination,
			       int x1,int y1,
			       int x0,int y0,int w,int h,
			       const Pixel*const* reference1,
			       int hdist1,int vdist1,
			       const Pixel*const* reference2,
			       int hdist2,int vdist2)
{
  Pixel a[16][16],b[16][16];
  Pixel* ap[16];
  Pixel* bp[16];
  for (int i=0;i<16;i++) { ap[i]=a[i]; bp[i]=b[i]; }

  MotionCompensateSingleSet(ap,0,0,x0,y0,w,h,reference1,hdist1,vdist1);
  MotionCompensateSingleSet(bp,0,0,x0,y0,w,h,reference2,hdist2,vdist2);

  for (int y=y1,yy=0;yy<16;y++,yy++)
    for (int x=x1,xx=0;xx<16;x++,xx++)
      { destination[y][x] = (((short)ap[yy][xx]) + bp[yy][xx] +1) /2; }
}


void MotionCompensateSingleSet(DiffPixel*const* destination,
			       int x1,int y1,
			       int x0,int y0,int w,int h,
			       const Pixel*const* reference, int hdist,int vdist)
{
  bool half_h = ((hdist&1)!=0);
  bool half_v = ((vdist&1)!=0);

  int inthdist = hdist/2;
  int intvdist = vdist/2;

  int xsign = Sign(hdist);
  int ysign = Sign(vdist);

  int yend = y1+h;

  intvdist = y0-y1+intvdist;
  inthdist = x0-x1+inthdist;

  if (!half_h && !half_v)
    {
      for (int y=y1;y<yend;y++)
	{
	  const Pixel*     rp = &reference[y+intvdist][inthdist+x1];
	        DiffPixel* dp = &destination[y][x1];

	  for (int i=w;i>0;i--)
	    *dp++ = *rp++;
	}
      return;
    }
  
  if (half_h && !half_v)
    {
      for (int y=y1;y<yend;y++)
	{
	  const Pixel*     rp1 = &reference[y+intvdist][inthdist+x1];
	  const Pixel*     rp2 = &reference[y+intvdist][inthdist+x1+xsign];
	        DiffPixel* dp  = &destination[y][x1];

	  for (int i=w;i>0;i--)
	    *dp++ = (*rp1++ + *rp2++ +1)>>1;
	}
      return;
    }

  if (!half_h && half_v)
    {
      for (int y=y1;y<yend;y++)
	{
	  const Pixel*     rp1 = &reference[y+intvdist      ][inthdist+x1];
	  const Pixel*     rp2 = &reference[y+intvdist+ysign][inthdist+x1];
	        DiffPixel* dp  = &destination[y][x1];

	  for (int i=w;i>0;i--)
	    *dp++ = (((short)*rp1++) + *rp2++ +1)>>1;
	}
      return;
    }

  if (half_h && half_v)
    {
      for (int y=y1;y<yend;y++)
	{
	  const Pixel*     rp1 = &reference[y+intvdist      ][inthdist+x1];
	  const Pixel*     rp2 = &reference[y+intvdist+ysign][inthdist+x1];
	  const Pixel*     rp3 = &reference[y+intvdist      ][inthdist+x1+xsign];
	  const Pixel*     rp4 = &reference[y+intvdist+ysign][inthdist+x1+xsign];
	        DiffPixel* dp  = &destination[y][x1];

	  for (int i=w;i>0;i--)
	    *dp++ = (((short)*rp1++) + *rp2++ + *rp3++ + *rp4++ + 2)>>2;
	}
      return;
    }
}


void MotionCompensateDoubleSet(DiffPixel*const* destination,
			       int x1,int y1,
			       int x0,int y0,int w,int h,
			       const Pixel*const* reference1,
			       int hdist1,int vdist1,
			       const Pixel*const* reference2,
			       int hdist2,int vdist2)
{
  Pixel a[16][16],b[16][16];
  Pixel* ap[16];
  Pixel* bp[16];
  for (int i=0;i<16;i++) { ap[i]=a[i]; bp[i]=b[i]; }

  MotionCompensateSingleSet(ap,0,0,x0,y0,w,h,reference1,hdist1,vdist1);
  MotionCompensateSingleSet(bp,0,0,x0,y0,w,h,reference2,hdist2,vdist2);

  for (int y=y1,yy=0;yy<16;y++,yy++)
    for (int x=x1,xx=0;xx<16;x++,xx++)
      { destination[y][x] = (((short)ap[yy][xx]) + bp[yy][xx] +1) /2; }
}














#if 0


void Decoder::Reset()
{
  m_current_img = NULL;
}

void Decoder::SetSequenceHeader(const SequenceHeaderData& sh)
{
  d_img1.CreateImage(YCrCb, sh.m_width, sh.m_height);
  d_img2.CreateImage(YCrCb, sh.m_width, sh.m_height);
  d_img3.CreateImage(YCrCb, sh.m_width, sh.m_height);
  d_chroma_format = sh.m_chroma_format;

  m_past_img    = &d_img1;
  m_future_img  = &d_img2;
  m_current_img = NULL;

  switch (d_chroma_format)
    {
    case CHROMA_420: d_chromaxfact= 8; d_chromayfact= 8;
                     d_chromaxdiv = 2; d_chromaydiv = 2; break;
    case CHROMA_422: d_chromaxfact= 8; d_chromayfact=16;
                     d_chromaxdiv = 2; d_chromaydiv = 1; break;
    case CHROMA_444: d_chromaxfact=16; d_chromayfact=16;
                     d_chromaxdiv = 2; d_chromaydiv = 1; break;
    }
}







  
  /* Dequantisierungsmatrizen vorbereiten */

  if (codedimg.m_pichdr.m_q_scale_type != d_q_scale_type ||
      codedimg.m_pichdr.m_intra_dc_precision != d_intra_dc_precision)
    {
      d_q_scale_type = codedimg.m_pichdr.m_q_scale_type;
      d_intra_dc_precision = codedimg.m_pichdr.m_intra_dc_precision;

      {for (int q=1;q<32;q++)
	for (int v=0;v<8;v++)
	  for (int u=0;u<8;u++)
	    {
	      qW[0][0][q][v][u] = QuantMatrixIntra[v][u] *
                                    qcode2qscale[codedimg.m_pichdr.m_q_scale_type][q];
	      qW[0][1][q][v][u] = QuantMatrixIntra[v][u] *
                                    qcode2qscale[codedimg.m_pichdr.m_q_scale_type][q];
	      qW[1][0][q][v][u] = 16 * qcode2qscale[codedimg.m_pichdr.m_q_scale_type][q];
	      qW[1][1][q][v][u] = 16 * qcode2qscale[codedimg.m_pichdr.m_q_scale_type][q];
	    }
      }

      {for (int q=1;q<32;q++)
	{
	  qW[0][0][q][0][0] = (1<<(11-codedimg.m_pichdr.m_intra_dc_precision));
	  qW[0][1][q][0][0] = (1<<(11-codedimg.m_pichdr.m_intra_dc_precision));
	}}
    }



  // Bild decodieren

#endif
