/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "coding/frameclass.hh"
#include "control/datapool.hh"
#include "mpeg/constants.hh"
#include "options.hh"


Task_ClassifyFrames_Regular::Task_ClassifyFrames_Regular(int firstframe,int nFrames)
{
  m_firstframe=firstframe;
  m_nFrames=nFrames;
  m_LastFrames=false;
}


void Task_ClassifyFrames_Regular::Execute(ProcessData& data)
{
  int last_frame = m_firstframe+m_nFrames-1;

  if (options.ProgressDisplay >= Options::ProgressDisplay_Normal)
    {
      ShowTaskInformation("frametype (regular)",data.threadnr);
      cerr << "frame " << m_firstframe << " to " << last_frame << endl;
      EndShowTaskInformation();
    }

  // Frame-Typ zuweisen

  for (int i=m_firstframe;i<=last_frame;i++)
    {
      // cout << "Get " << i << " -----------\n";
      ImageData* img = ImageData::GetNr(i);
      assert(img);

      if (i<=m_IDistance-m_PDistance) // erste GOP
        img->pichdr.m_temporal_reference = i;
      else
        img->pichdr.m_temporal_reference = (i+m_PDistance-1)%m_IDistance;

      if (options.ProgressDisplay >= Options::ProgressDisplay_Maximal)
        cout << "Frame " << i << " (temp.ref " << img->pichdr.m_temporal_reference << ") is a : ";

      if (((i+m_PDistance-1)%m_IDistance)==0 || i==0)
        {
          if (options.ProgressDisplay >= Options::ProgressDisplay_Maximal) cout << "[GOP] ";
        }

      if ((i%m_IDistance)==0)
        {
          img->pichdr.m_picture_coding_type = PICTYPE_I;
          img->SetGOPBeginning();
          img->IsClosedGOP();
          if (options.ProgressDisplay >= Options::ProgressDisplay_Maximal) cout << "I\n";
        }
      else if ((i%m_PDistance)==0)
        {
          img->pichdr.m_picture_coding_type = PICTYPE_P;
          if (options.ProgressDisplay >= Options::ProgressDisplay_Maximal) cout << "P\n";
        }
      else
        {
          if (m_LastFrames && i==last_frame)
            {
              img->pichdr.m_picture_coding_type = PICTYPE_P;
              if (options.ProgressDisplay >= Options::ProgressDisplay_Maximal) cout << "(P)\n";
            }
          else
            {
              img->pichdr.m_picture_coding_type = PICTYPE_B;
              if (options.ProgressDisplay >= Options::ProgressDisplay_Maximal) cout << "B\n";
            }
        }
    }

  m_last_frame_type_assigned = last_frame;
}








Task_ClassifyFrames_IOnly::Task_ClassifyFrames_IOnly(int firstframe,int nFrames)
{
  m_firstframe=firstframe;
  m_nFrames=nFrames;
  m_LastFrames=false;
}


void Task_ClassifyFrames_IOnly::Execute(ProcessData& data)
{
  int last_frame = m_firstframe+m_nFrames-1;   // m_nFrames/2 -> Ein bischen stressen.

  if (options.ProgressDisplay >= Options::ProgressDisplay_Normal)
    {
      ShowTaskInformation("frametype (I only)",data.threadnr);
      cerr << "frame " << m_firstframe << " to " << last_frame << endl;
      EndShowTaskInformation();
    }

  for (int i=m_firstframe;i<=last_frame;i++)
    {
      ImageData* img = ImageData::GetNr(i);
      assert(img);

      if ((m_firstframe % options.framepattern.GOPSize)==0)
        {
          img->SetGOPBeginning();
          img->IsClosedGOP();
        }

      if (options.ProgressDisplay >= Options::ProgressDisplay_Maximal)
        cerr << "Frame " << i << " is a : I\n";

      img->pichdr.m_picture_coding_type = PICTYPE_I;
    }

  m_last_frame_type_assigned = last_frame;
}
