/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include <coding/motionest.hh>


void MotionEstimator::AdjustRange(const SearchParam& spar,
                                  int x,int y, int width,int height,
                                  int& left,int& right,int& top,int& bottom) const
{
  left   = spar.minx;
  right  = spar.maxx;
  top    = spar.miny;
  bottom = spar.maxy;

  if (x+left<0) left=-x;
  if (y+top <0) top =-y;

  if (x+spar.blk_width +right >width ) right =width -x-spar.blk_width;
  if (y+spar.blk_height+bottom>height) bottom=height-y-spar.blk_height;
}



bool BlockComparator::Error(BCData* bcdata,const ComparePos& cmppos,
                            uint32& error,uint32 maxerror,int subfact) const
{
  ErrorMulti(bcdata,1,&cmppos,&error,maxerror,subfact);
  return true;
}


void BlockComparator::ErrorMulti(BCData* bcdata,int nVecs,const ComparePos* cmppos,
                                 uint32* error,uint32 maxerror,int subfact) const
{
  for (int i=0;i<nVecs;i++)
    Error(bcdata,cmppos[i],error[i],maxerror,subfact);
}




#include "coding/bcstd.hh"

#if USE_MMX
#include "coding/bcmmx.hh"
#include "arch/MMX/mmx.hh"
#elif USE_VIS
#include "coding/bcVIS.hh"
#include "arch/VIS/vischk.hh"
#endif


BlockComparator* BlockComparator::CreateDefaultBC()
{
#if USE_VIS
  if (VIS_Supported)
    return new BlockComparator_VIS;
#elif USE_MMX
  if (MMX_Supported)
    return new BlockComparator_MMX;
#endif
  return new BlockComparator_Standard;
}
