/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "coding/motionest_full.hh"
#include <math.h>
#include <iostream.h>
#include <fstream.h>
#include <stdio.h>
#include "global.hh"

#define LOG_OUTPUT 0

struct MEData_FullSearch : public MEData
{
  ~MEData_FullSearch() { delete bcdata; }

  BCData*     bcdata;
  SearchParam sparam;
};


MotionEstimator_Full::MotionEstimator_Full()
{
  d_blkcmp = BlockComparator::CreateDefaultBC();
}

MotionEstimator_Full::~MotionEstimator_Full()
{
  delete d_blkcmp;
}


MEData* MotionEstimator_Full::CreateMEData(const SearchParam& spar,const MEData*)
{
  MEData_FullSearch* medata = new MEData_FullSearch;
  medata->bcdata = d_blkcmp->CreateBCData(spar);
  medata->sparam = spar;

  return medata;
}

void MotionEstimator_Full::SetSearchParams(MEData* medata, const Bitmap<Pixel>* bitmap,
                                           bool frame, bool top)
{
  MEData_FullSearch* med = dynamic_cast<MEData_FullSearch*>(medata);
  assert(med);

  c_medata = med;
  c_bcdata = med->bcdata;
  c_frame_flag = frame;
  c_top_flag   = top;
  c_img_width  = bitmap->AskWidth();
  c_img_height = bitmap->AskHeight();
  c_pixels_per_block = med->sparam.blk_width * med->sparam.blk_height;

  d_blkcmp->SetImageBitmap(*bitmap,frame,top);
}


bool MotionEstimator_Full::Estimate(int x,int y, int& mx,int& my,
					   double& error,double upper_error_bound)
{
#if LOG_OUTPUT
  char buf[100];
  sprintf(buf,"/usr/home/dirk/mv/mv%03d_%02d_%02d.dat",globalstats.AskCurrentFrameNr(),y/16,x/16);
  cout << " WRITETO: " << buf << endl;
  ofstream ostr(buf);

  mx = -1000;
#endif

  MV_COMMON_PART
  

  uint32 maxerror = (uint32)(upper_error_bound*c_pixels_per_block);
  uint32 err;

  int bestmv_x=0,bestmv_y=0;
  const int step = 1;  // <<<<<<<<<<<<<<<<<<<<<< Performance-Hack
  bool mv_found = false;

#if 1
  for (int moty=top; moty<=bottom; moty+=step)
    {
      for (int motx=left; motx<=right; motx+=step)
	{
	  cmppos.ref_x = x0+motx;
	  cmppos.ref_y = y0+moty;
	  
	  if (d_blkcmp->Error(c_bcdata,cmppos,err,LOG_OUTPUT ? MASK31 : maxerror))
	    {
#if LOG_OUTPUT
	      ostr << motx << ' ' << -moty << ' ' << err << ' ' << maxerror << endl;
#endif
	      
	      if (err<maxerror)
		{
		  maxerror = err;
		  bestmv_x = motx;
		  bestmv_y = moty;
		  mv_found = true;
		}
	    }
	}
#if LOG_OUTPUT
      ostr << endl;
#endif
    }
#else

  for (int y=0;-y >= top && y<=bottom; y++)
    {
      if (y<=bottom)
        {
          for (int motx=left; motx<=right; motx+=step)
            {
              cmppos.ref_x = x0+motx;
              cmppos.ref_y = y0+y;
              
              if (d_blkcmp->Error(c_bcdata,cmppos,err,LOG_OUTPUT ? MASK31 : maxerror))
                {
                  if (err<maxerror)
                    {
                      maxerror = err;
                      bestmv_x = motx;
                      bestmv_y = y;
                      mv_found = true;
                    }
                }
            }
        }

      if (-y>=top && y>0)
        {
          for (int motx=left; motx<=right; motx+=step)
            {
              cmppos.ref_x = x0+motx;
              cmppos.ref_y = y0-y;
              
              if (d_blkcmp->Error(c_bcdata,cmppos,err,LOG_OUTPUT ? MASK31 : maxerror))
                {
                  if (err<maxerror)
                    {
                      maxerror = err;
                      bestmv_x = motx;
                      bestmv_y = -y;
                      mv_found = true;
                    }
                }
            }
        }
    }
#endif
  
  if (mv_found)
    {
      mx = bestmv_x;
      my = bestmv_y;
      error = ((double)maxerror)/c_pixels_per_block;
    }

  return mv_found || nomv_vector_valid;
}
