/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#ifndef CODING_ME1_SIMPLE_HH
#define CODING_MOTIONEST_HH

#include "types.hh"
#include "image/bitmap.hh"
#include <stdlib.h>


struct SearchParam
{
  int minx,maxx;
  int miny,maxy;

  bool halfpel;

  int blk_width,blk_height;

  const Bitmap<Pixel>* bitmap;
};


struct MVData
{
  virtual ~MVData() { }
};


class MotionEstimator
{
public:
  MotionEstimator() { }
  virtual ~MotionEstimator() { }


  MVData* CreateMVData(const SearchParam&);

  virtual void SetImageBitmap(const Pixel*const*) = 0;
  virtual bool Estimate(int x,int y, int& mx,int& my,
                        double& error,double upper_error_bound=512.0);

protected:
  struct ComparePos
  {
    int img_x,img_y;  // Linke obere Ecke im der Originalbild-Bitmap, bei der die Suche beginnen soll.
    int end_x,end_y;  // Rechte untere Position, bei der die Suche abgebrochen werden soll.

    int ref_x,ref_y;  // Linke obere Ecke in der Referenz-Bitmap, bei der die Suche beginnen soll.
  };

  void AdjustRange(int x,int y,int& left,int& right,int& top,int& bottom) const;
  bool Error(MVData_StdBitmaps,
             const ComparePos&,  // Positionen der 16x16 Bloecke und Half-Pel-Flag
	     uint32 maxerror,    // Obere Grenze fuer den Fehler.
	                         // Bei Ueberschreitung kann abgebrochen werden.
	     uint32& error,      // Rueckgabe des Gesamtfehlers.
	     int subfact=1       // Unterabtastungsfaktor
	     ) const;

  bool Error_Simple(const ComparePos&,uint32 maxerror,uint32& error,int subfact=1) const;
  bool Error_MMX   (const ComparePos&,uint32 maxerror,uint32& error,int subfact=1) const;
  //bool Error_VIS   (const ComparePos&,uint32 maxerror,uint32& error,int subfact=1) const;
};


#endif
