/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "coding/motionest_full.hh"
#include <math.h>
#include <iostream.h>
#include <fstream.h>
#include <stdio.h>
#include "global.hh"

#define LOG_OUTPUT 0

// inline uint32 min(uint32 a,uint32 b) { if (a<=b) return a; else return b; }

bool MotionEstimator_Full::Estimate(int x,int y, int& mx,int& my,double& error,double upper_error_bound)
{
  int left,right,top,bottom;
  AdjustRange(x,y, left,right,top,bottom);

  ComparePos cmppos;
  cmppos.img_x = x;
  cmppos.img_y = y;
  cmppos.end_x = x+d_pattern_width;
  cmppos.end_y = y+d_pattern_height;

  uint32 maxerror = (uint32)(upper_error_bound*d_pixels_per_block);
  uint32 err;

  bool mv_found      = false;
  bool nomv_possible = false;


#if LOG_OUTPUT
	char buf[100];
	sprintf(buf,"/usr/home/dirk/mv/mv%03d_%02d_%02d.dat",globalstats.AskCurrentFrameNr(),y/16,x/16);
        cout << " WRITETO: " << buf << endl;
	ofstream ostr(buf);
#endif

  // MAE-Fehler berechnen, fuer den Fall, dass der MV nicht veraendert wird.

  if (mx>=left && mx<=right && my>=top && my<=bottom)
    {
      cmppos.ref_x = x*(d_halfpel ? 2 : 1)+mx;
      cmppos.ref_y = y*(d_halfpel ? 2 : 1)+my;

      uint32 error0;
      if (Error(cmppos,maxerror,error0))
	{
	  double nomv_error = ((double)error0)/d_pixels_per_block;
	  error=nomv_error;

	  /* Wenn der Fehler sehr gering ist, dann gleich diesen Vektor nehmen und nicht
	     weitersuchen. */
	  if (nomv_error <= 1.0)
	    {
	      return true;
	    }
	  else if (nomv_error <= 3.0)
	    {
	      maxerror = error0 /2;
	    }
	  else
	    {
	      maxerror = error0 *29/32;  // *0.9
	    }

	  nomv_possible = true;
	}
    }


  // MVs durchsuchen

  int bestmv_x=0,bestmv_y=0;

  const int step = 1;  // <<<<<<<<<<<<<<<<<<<<<< Performance-Hack

  int x0 = x*(d_halfpel ? 2 : 1);
  int y0 = y*(d_halfpel ? 2 : 1);

  for (int moty=top; moty<=bottom; moty+=step)
    {
      for (int motx=left; motx<=right; motx+=step)
	{
	  cmppos.ref_x = x0+motx;
	  cmppos.ref_y = y0+moty;
	  
	  if (Error(cmppos,LOG_OUTPUT ? MASK31 : maxerror,err))
	    {
#if LOG_OUTPUT
	      ostr << motx << ' ' << -moty << ' ' << err << ' ' << maxerror << endl;
#endif
	      
	      if (err<maxerror)
		{
		  maxerror = err;
		  bestmv_x = motx;
		  bestmv_y = moty;
		  mv_found = true;
		}
	    }
	}
#if LOG_OUTPUT
      ostr << endl;
#endif
    }
  
  if (mv_found)
    {
      mx = bestmv_x;
      my = bestmv_y;
      error = ((double)maxerror)/d_pixels_per_block;
    }

  return mv_found || nomv_possible;
}
