/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include <coding/motionest.hh>
#include <iostream.h>
#ifdef USE_MMX
#include "arch/mmx.hh"
#endif

#define MYVERSION 0
#define INTEL_ORIGINAL 0

/*
  Tagesschau-Film, Codierung der ersten 50 Bilder mit FullSearch.   (-F -e 50)

                     AMD k6 233
  Meine Version:       183.06s
  Intel original:      190.30s
  Intel modifiziert:   182.74s
 */


struct MVData_StdBitmaps : public MVData
{
  MVData_StdBitmaps()
    : d_reference(NULL),
      d_reference_halfh(NULL),
      d_reference_halfv(NULL),
      d_reference_halfhv(NULL) { }
  ~MVData_StdBitmaps();

  double d_pixels_per_block;

  inline const MVPixel*const* AskPixels(int& x,int& y) const;

private:
      MVPixel*const * d_reference;          // Original-Bitmap (1:1)
      MVPixel*const * d_reference_halfh;    // Bitmap aus jeweils 2 horiz. gemittelten Pixeln
      MVPixel*const * d_reference_halfv;    // Bitmap aus jeweils 2 verti. gemittelten Pixeln
      MVPixel*const * d_reference_halfhv;   // Bitmap aus jeweils 4 gemittelten Pixeln
  const Pixel*const * d_image;

  bool d_halfpel;

#ifndef MVPIX_EQ_PIX
  Bitmap<MVPixel> d_reference_bm;
#endif
  Bitmap<MVPixel> d_reference_bm_h;
  Bitmap<MVPixel> d_reference_bm_v;
  Bitmap<MVPixel> d_reference_bm_hv;
};


inline const MVPixel*const* MVData_StdBitmaps::AskPixels(int& x,int& y) const
{
  if (!d_halfpel)
    return d_reference;

  const MVPixel*const* ref = d_reference;

  bool halfh = (x & 1);
  bool halfv = (y & 1);

  x >>= 1;
  y >>= 1;

  if (halfv)
    {
      if (halfh)
        ref = d_reference_halfhv;
      else
        ref = d_reference_halfv;
    }
  else
    {
      if (halfh)
        ref = d_reference_halfh;
    }

  return ref;
}





MotionEstimator::MotionEstimator()
{
}

MotionEstimator::~MotionEstimator()
{
}

void MotionEstimator::SetReferenceBitmap(const Bitmap<Pixel>& bitmap,bool halfpel)
{
  const int w = d_width  = bitmap.AskWidth();
  const int h = d_height = bitmap.AskHeight();

  const Pixel*const* refbm = bitmap.AskFrame_const();

  d_halfpel = halfpel;


  /* Wenn schon die HalfPel-Bitmaps erzeugt wurden, dann muessen sie
     eigentlich schon die richtige Groesse haben. */

  if (( d_halfpel && d_reference_halfh) ||
      (!d_halfpel && d_reference))
    {
      if (d_halfpel)
        {
          assert(d_reference_bm_h. AskWidth()  == w-1);
          assert(d_reference_bm_h. AskHeight() == h  );
          assert(d_reference_bm_v. AskWidth()  == w  );
          assert(d_reference_bm_v. AskHeight() == h-1);
          assert(d_reference_bm_hv.AskWidth()  == w-1);
          assert(d_reference_bm_hv.AskHeight() == h-1);
        }
    }
  else
    {
#ifndef MVPIX_IS_PIX
      if (!d_reference)
        {
          d_reference_bm.CreateBitmap(w,h);
          d_reference = d_reference_bm.AskFrame();
        }
#endif

      if (d_halfpel)
        {
          // Half-Pel-Bitmaps neu erzeugen

          d_reference_bm_h .CreateBitmap(w-1,h  );
          d_reference_bm_v .CreateBitmap(w  ,h-1);
          d_reference_bm_hv.CreateBitmap(w-1,h-1);

          d_reference_halfh  = d_reference_bm_h .AskFrame();
          d_reference_halfv  = d_reference_bm_v .AskFrame();
          d_reference_halfhv = d_reference_bm_hv.AskFrame();
        }
    }


  // --- Neue Bitmaps fuellen ---

#ifndef MVPIX_IS_PIX
  {for (int y=0;y<h;y++)
    for (int x=0;x<w;x++)
      d_reference[y][x] = refbm[y][x];}
#else
  d_reference = const_cast<Pixel*const*>(refbm);
#endif

  if (d_halfpel)
    {
      {for (int y=0;y<h;y++)
        for (int x=0;x<w-1;x++)
          d_reference_halfh[y][x] = (refbm[y][x] + refbm[y][x+1] +1)/2;}

      {for (int y=0;y<h-1;y++)
        for (int x=0;x<w;x++)
          d_reference_halfv[y][x] = (refbm[y][x] + refbm[y+1][x] +1)/2;}

      {for (int y=0;y<h-1;y++)
        for (int x=0;x<w-1;x++)
          d_reference_halfhv[y][x] = (refbm[y  ][x] + refbm[y  ][x+1] +
                                      refbm[y+1][x] + refbm[y+1][x+1] +2)/4;}
    }
  }

void MotionEstimator::AdjustRange(int x,int y,int& left,int& right,int& top,int& bottom) const
{
  left   = d_range_minx;
  right  = d_range_maxx;
  top    = d_range_miny;
  bottom = d_range_maxy;

  if (x+left<0) left=-x;
  if (y+top <0) top =-y;

  if (x+d_pattern_width +right >d_width ) right =d_width -x-d_pattern_width;
  if (y+d_pattern_height+bottom>d_height) bottom=d_height-y-d_pattern_height;
}


bool MotionEstimator::Error(const ComparePos& cmppos,uint32 maxerror,uint32& error,
                            int subfact) const
{
#if USE_MMX
  if (MMX_Supported)
    return Error_MMX(cmppos,maxerror,error,subfact);
#endif

  return Error_Simple(cmppos,maxerror,error,subfact);
}


/*
  Gesamte mean absolute difference (MAD) zwischen zwei 16x16 Bloecken berechnen.
 */
bool MotionEstimator::Error_Simple(const ComparePos& cmppos,uint32 maxerror,uint32& error,
                                   int subfact) const
{
  const MVPixel*const* ref = d_reference;

  int refx,refy;

  if (d_halfpel)
    {
      bool halfh = (cmppos.ref_x & 1);
      bool halfv = (cmppos.ref_y & 1);

      refx = (cmppos.ref_x >> 1);
      refy = (cmppos.ref_y >> 1);

      if (halfv)
        {
          if (halfh)
            ref = d_reference_halfhv;
          else
            ref = d_reference_halfv;
        }
      else
        {
          if (halfh)
            ref = d_reference_halfh;
        }
    }
  else
    {
      refx = cmppos.ref_x;
      refy = cmppos.ref_y;
    }

  error=0;

  for (int iy=cmppos.img_y , ry=refy ;
       iy<cmppos.end_y ;
       iy+=subfact,ry+=subfact)
    {
#if INTERLEAVED_MVSEARCH
      const   Pixel* ip1 = &d_image[iy][cmppos.img_x];  // Zeiger ins Originalbild
      const MVPixel* rp1 = &ref    [ry][refx];          // Zeiger ins Referenzbild
      const MVPixel* ip2 = ip1 + 8*subfact;
      const MVPixel* rp2 = rp1 + 8*subfact;

      int diff1,diff2;
      int error2=0;

      diff1 = *ip1-*rp1; if (diff1>=0) error  += diff1; else error  -= diff1; ip1+=subfact; rp1+=subfact;
      diff2 = *ip2-*rp2; if (diff2>=0) error2 += diff2; else error2 -= diff2; ip2+=subfact; rp2+=subfact;
      diff1 = *ip1-*rp1; if (diff1>=0) error  += diff1; else error  -= diff1; ip1+=subfact; rp1+=subfact;
      diff2 = *ip2-*rp2; if (diff2>=0) error2 += diff2; else error2 -= diff2; ip2+=subfact; rp2+=subfact;
      diff1 = *ip1-*rp1; if (diff1>=0) error  += diff1; else error  -= diff1; ip1+=subfact; rp1+=subfact;
      diff2 = *ip2-*rp2; if (diff2>=0) error2 += diff2; else error2 -= diff2; ip2+=subfact; rp2+=subfact;
      diff1 = *ip1-*rp1; if (diff1>=0) error  += diff1; else error  -= diff1; ip1+=subfact; rp1+=subfact;
      diff2 = *ip2-*rp2; if (diff2>=0) error2 += diff2; else error2 -= diff2; ip2+=subfact; rp2+=subfact;
      diff1 = *ip1-*rp1; if (diff1>=0) error  += diff1; else error  -= diff1; ip1+=subfact; rp1+=subfact;
      diff2 = *ip2-*rp2; if (diff2>=0) error2 += diff2; else error2 -= diff2; ip2+=subfact; rp2+=subfact;
      diff1 = *ip1-*rp1; if (diff1>=0) error  += diff1; else error  -= diff1; ip1+=subfact; rp1+=subfact;
      diff2 = *ip2-*rp2; if (diff2>=0) error2 += diff2; else error2 -= diff2; ip2+=subfact; rp2+=subfact;
      diff1 = *ip1-*rp1; if (diff1>=0) error  += diff1; else error  -= diff1; ip1+=subfact; rp1+=subfact;
      diff2 = *ip2-*rp2; if (diff2>=0) error2 += diff2; else error2 -= diff2; ip2+=subfact; rp2+=subfact;
      diff1 = *ip1-*rp1; if (diff1>=0) error  += diff1; else error  -= diff1; ip1+=subfact; rp1+=subfact;
      diff2 = *ip2-*rp2; if (diff2>=0) error2 += diff2; else error2 -= diff2; ip2+=subfact; rp2+=subfact;

      error += error2;
#else
      const   Pixel* ip = &d_image    [iy][cmppos.img_x];  // Zeiger ins Originalbild
      const MVPixel* rp = &ref        [ry][refx];  // Zeiger ins Referenzbild

      int diff;

      diff = *ip - *rp; if (diff>=0) error += diff; else error -= diff;    ip+=subfact; rp+=subfact;
      diff = *ip - *rp; if (diff>=0) error += diff; else error -= diff;    ip+=subfact; rp+=subfact;
      diff = *ip - *rp; if (diff>=0) error += diff; else error -= diff;    ip+=subfact; rp+=subfact;
      diff = *ip - *rp; if (diff>=0) error += diff; else error -= diff;    ip+=subfact; rp+=subfact;
      diff = *ip - *rp; if (diff>=0) error += diff; else error -= diff;    ip+=subfact; rp+=subfact;
      diff = *ip - *rp; if (diff>=0) error += diff; else error -= diff;    ip+=subfact; rp+=subfact;
      diff = *ip - *rp; if (diff>=0) error += diff; else error -= diff;    ip+=subfact; rp+=subfact;
      diff = *ip - *rp; if (diff>=0) error += diff; else error -= diff;    ip+=subfact; rp+=subfact;
      diff = *ip - *rp; if (diff>=0) error += diff; else error -= diff;    ip+=subfact; rp+=subfact;
      diff = *ip - *rp; if (diff>=0) error += diff; else error -= diff;    ip+=subfact; rp+=subfact;
      diff = *ip - *rp; if (diff>=0) error += diff; else error -= diff;    ip+=subfact; rp+=subfact;
      diff = *ip - *rp; if (diff>=0) error += diff; else error -= diff;    ip+=subfact; rp+=subfact;
      diff = *ip - *rp; if (diff>=0) error += diff; else error -= diff;    ip+=subfact; rp+=subfact;
      diff = *ip - *rp; if (diff>=0) error += diff; else error -= diff;    ip+=subfact; rp+=subfact;
      diff = *ip - *rp; if (diff>=0) error += diff; else error -= diff;    ip+=subfact; rp+=subfact;
      diff = *ip - *rp; if (diff>=0) error += diff; else error -= diff;    ip+=subfact; rp+=subfact;
#endif

#if 1
      if (error>maxerror)
	return false;
#endif
    }

  if (subfact!=1)
    error *= subfact*subfact;

  return true;
}


typedef unsigned long long mmx;

bool MotionEstimator::Error_MMX(const ComparePos& cmppos,uint32 maxerror,uint32& error,
                                int subfact) const
{
  const MVPixel*const* ref = d_reference;

  int refx,refy;

  if (d_halfpel)
    {
      bool halfh = (cmppos.ref_x & 1);
      bool halfv = (cmppos.ref_y & 1);

      refx = (cmppos.ref_x >> 1);
      refy = (cmppos.ref_y >> 1);

      if (halfv)
        {
          if (halfh)
            ref = d_reference_halfhv;
          else
            ref = d_reference_halfv;
        }
      else
        {
          if (halfh)
            ref = d_reference_halfh;
        }
    }
  else
    {
      refx = cmppos.ref_x;
      refy = cmppos.ref_y;
    }

  mmx diff=0;
  error=0;

  __asm__ __volatile__
    (
#if MYVERSION
     "pxor      %mm3,%mm3\n\t"    // mm3 auf 0 setzen
     "pxor      %mm4,%mm4\n\t"    // Fehler auf 0 setzen (mm4)
#else
     "pxor      %mm6,%mm6\n\t"    // mm6 auf 0 setzen
     "pxor      %mm7,%mm7\n\t"    // Fehler auf 0 setzen (mm4)    (Intel-Code)
#endif
     );

  for (int iy=cmppos.img_y , ry=refy ;
       iy<cmppos.end_y ;
       iy+=subfact,ry+=subfact)
    {
      const mmx* ip = (const mmx*)&d_image[iy][cmppos.img_x];  // Zeiger ins Originalbild
      const mmx* rp = (const mmx*)&ref    [ry][refx];          // Zeiger ins Referenzbild

      mmx pa1,pa2,pb1,pb2;
      pa1 = *ip++;
      pa2 = *ip;
      pb1 = *rp++;
      pb2 = *rp;

#if MYVERSION
      // Meine Version

      __asm__ __volatile__
        (
         // Die ersten 8 Pixel

         "movq      %0,%%mm0\n\t"       // Daten A -> mm0
         "movq      %1,%%mm1\n\t"       // Daten B -> mm1
         "movq      %%mm0,%%mm2\n\t"    // Daten A -> mm2   (Kopie)
         "psubusb   %%mm1,%%mm2\n\t"    // A-B     -> mm2
         "psubusb   %%mm0,%%mm1\n\t"    // B-A     -> mm1
         "por       %%mm2,%%mm1\n\t"    // Zusammenfassen der Differenzen -> mm1
         "movq      %%mm1,%%mm2\n\t"    //                       und nach -> mm2
         "punpckhbw %%mm3,%%mm1\n\t"    // Oberen  4 Werte auf 16bit erweitert nach -> mm1
         "punpcklbw %%mm3,%%mm2\n\t"    // Unteren 4 Werte auf 16bit erweitert nach -> mm2
         "paddw     %%mm1,%%mm4\n\t"    // mm1 auf alten Fehler addieren
         "paddw     %%mm2,%%mm4\n\t"    // mm2 auf alten Fehler addieren

         // Die zweiten 8 Pixel

         "movq      %2,%%mm0\n\t"       // Daten A -> mm0
         "movq      %3,%%mm1\n\t"       // Daten B -> mm1
         "movq      %%mm0,%%mm2\n\t"    // Daten A -> mm2   (Kopie)
         "psubusb   %%mm1,%%mm2\n\t"    // A-B     -> mm2
         "psubusb   %%mm0,%%mm1\n\t"    // B-A     -> mm1
         "por       %%mm2,%%mm1\n\t"    // Zusammenfassen der Differenzen -> mm1
         "movq      %%mm1,%%mm2\n\t"    //                       und nach -> mm2
         "punpckhbw %%mm3,%%mm1\n\t"    // Oberen  4 Werte auf 16bit erweitert nach -> mm1
         "punpcklbw %%mm3,%%mm2\n\t"    // Unteren 4 Werte auf 16bit erweitert nach -> mm2
         "paddw     %%mm1,%%mm4\n\t"    // mm1 auf alten Fehler addieren
         "paddw     %%mm2,%%mm4\n\t"    // mm2 auf alten Fehler addieren
         : : "m" (pa1), "m" (pb1), "m" (pa2), "m" (pb2)
         );
#else

#if INTEL_ORIGINAL
      // Intel original

      __asm__ __volatile__
        (
         "movq      %0,%%mm0\n\t"
         "movq      %1,%%mm1\n\t"
         "movq      %%mm0,%%mm2\n\t"
         "movq      %2,%%mm3\n\t"
         "psubusb   %%mm1,%%mm0\n\t"
         "psubusb   %%mm2,%%mm1\n\t"
         "movq      %3,%%mm4\n\t"
         "por       %%mm1,%%mm0\n\t"
         "movq      %%mm3,%%mm5\n\t"
         "movq      %%mm0,%%mm1\n\t"
         "psubusb   %%mm4,%%mm3\n\t"
         "punpcklbw %%mm6,%%mm0\n\t"
         "psubusb   %%mm5,%%mm4\n\t"
         "psrlq     $32,%%mm1\n\t"
         "por       %%mm4,%%mm3\n\t"
         "punpcklbw %%mm6,%%mm1\n\t"
         "movq      %%mm3,%%mm4\n\t"
         "punpcklbw %%mm6,%%mm3\n\t"
         "paddw     %%mm0,%%mm7\n\t"
         "psrlq     $32,%%mm4\n\t"
         "paddw     %%mm1,%%mm7\n\t"
         "punpcklbw %%mm6,%%mm4\n\t"
         "paddw     %%mm3,%%mm7\n\t"
         "paddw     %%mm4,%%mm7\n\t"
         : : "m" (pa1), "m" (pb1), "m" (pa2), "m" (pb2)
         );
#else
      // Intel modifiziert

      __asm__ __volatile__
        (
         "movq      %0,%%mm0\n\t"
         "movq      %1,%%mm1\n\t"
         "movq      %%mm0,%%mm2\n\t"
         "movq      %2,%%mm3\n\t"
         "psubusb   %%mm1,%%mm0\n\t"
         "psubusb   %%mm2,%%mm1\n\t"
         "movq      %3,%%mm4\n\t"
         "por       %%mm1,%%mm0\n\t"
         "movq      %%mm3,%%mm5\n\t"
         "movq      %%mm0,%%mm1\n\t"
         "psubusb   %%mm4,%%mm3\n\t"
         "punpcklbw %%mm6,%%mm0\n\t"
         "psubusb   %%mm5,%%mm4\n\t"
         "por       %%mm4,%%mm3\n\t"
         "punpckhbw %%mm6,%%mm1\n\t"
         "movq      %%mm3,%%mm4\n\t"
         "punpcklbw %%mm6,%%mm3\n\t"
         "paddw     %%mm0,%%mm7\n\t"
         "paddw     %%mm1,%%mm7\n\t"
         "punpckhbw %%mm6,%%mm4\n\t"
         "paddw     %%mm3,%%mm7\n\t"
         "paddw     %%mm4,%%mm7\n\t"
         : : "m" (pa1), "m" (pb1), "m" (pa2), "m" (pb2)
         );
#endif

#endif

#if 0
      if (diff & 0xFFFF > maxerror/2)
        {
          error  =  diff      & 0xFFFF;
          error += (diff>>16) & 0xFFFF;
          error += (diff>>32) & 0xFFFF;
          error += (diff>>48) & 0xFFFF;
          if (error>maxerror)
            return false;
        }
#endif

      /*
      printf("Data: %016llx / %016llx\n", pa1,pa2);
      printf("Data: %016llx / %016llx\n", pb1,pb2);
      printf("Diff: %016llx\n", diff);

      exit(0);
      */
    }

  __asm__ __volatile__
    (
#if MYVERSION
     "movq    %%mm4,%0\n\t"
#else
     "movq    %%mm7,%0\n\t"
#endif
     "emms\n\t"
     : "=m" (diff)
    );

  error  =  diff      & 0xFFFF;
  error += (diff>>16) & 0xFFFF;
  error += (diff>>32) & 0xFFFF;
  error += (diff>>48) & 0xFFFF;

  if (subfact!=1)
    error *= subfact*subfact;

  if (error>maxerror)
    return false;
  else
    return true;
}


#include "image/bitmap.cc"
