/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "coding/taskmvsearch.hh"
#include "control/datapool.hh"
#include "options.hh"

void Task_MVSearch_0::Execute(ProcessData& data)
{
  ImageData* img = ImageData::GetNr(m_framenr);
  assert(img);

  if (options.ProgressDisplay >= Options::ProgressDisplay_Normal)
    {
      ShowTaskInformation("MV search 0/0",data.threadnr);
      cerr << "frame " << m_framenr << endl;
      EndShowTaskInformation();
    }

  for (int y=0;y<img->mblks.AskHeight();y++)
    for (int x=0;x<img->mblks.AskWidth();x++)
      {
        Macroblock& mb = img->mblks.Ask(y,x);
        if (m_forward)
          { mb.m_forward1.m_habs=0;   mb.m_forward1.m_vabs=0;  }
        else
          { mb.m_backward1.m_habs=0;  mb.m_backward1.m_vabs=0; }
      }
}


void Task_MVSearch::Execute(ProcessData& data)
{
  ImageData* img = ImageData::GetNr(m_framenr);
  assert(img);

  if (options.ProgressDisplay >= Options::ProgressDisplay_Normal)
    {
      ShowTaskInformation("MV search",data.threadnr);
      if (m_forward) cerr << "->"; else cerr << "<-";
      cerr << " frame " << m_framenr << endl;
      EndShowTaskInformation();
    }

  bool EnableDisplay = (options.ProgressDisplay >= Options::ProgressDisplay_Maximal);

  // forward-MVs berechnen

  MotionEstimator* mest = img->GetMotionEstimator();
  assert(mest);

  if (m_forward)
    {
      assert(img->prev_ip);
      assert(img->prev_ip->medata);
      mest->SetSearchParams((class MEData*)img->prev_ip->medata, &img->source->AskBitmap(Bitmap_Y),true);
    }
  else
    {
      assert(img->next_ip);
      assert(img->next_ip->medata);
      mest->SetSearchParams((class MEData*)img->next_ip->medata, &img->source->AskBitmap(Bitmap_Y),true);
    }

  int mv_h,mv_v;

  for (int y=0;y<img->mblks.AskHeight();y++)
    {
      mv_h = 0;
      mv_v = 0;

      for (int x=0;x<img->mblks.AskWidth();x++)
        {
          Macroblock& mb = img->mblks.Ask(y,x);

          double err;
          bool block_found = mest->Estimate(x*16,y*16,mv_h,mv_v,err);
          assert(block_found);
          //cout << "                         " << mv_h << "  " << mv_v << endl;

          if (m_forward)
            {
              mb.m_forward1.m_habs=mv_h;
              mb.m_forward1.m_vabs=mv_v;
            }
          else
            {
              mb.m_backward1.m_habs=mv_h;
              mb.m_backward1.m_vabs=mv_v;
            }
        }
      
      if (EnableDisplay) cerr << '.' << y;
    }
  if (EnableDisplay) cerr << endl;

  img->GiveBackMotionEstimator(mest);
}
