/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "coding/varianz.hh"
#include "control/datapool.hh"
#include "options.hh"


static double lastmean,lastvar;

Task_ComputeVarianzMean::Task_ComputeVarianzMean(int framenr)
  : m_framenr(framenr)
{
}


void Task_ComputeVarianzMean::Execute(ProcessData& data)
{
  ImageData* img = ImageData::GetNr(m_framenr);
  assert(img);

  if (options.ProgressDisplay >= Options::ProgressDisplay_Normal)
    {
      ShowTaskInformation("compute Var/Mean",data.threadnr);
      cerr << "frame " << m_framenr << endl;
      EndShowTaskInformation();
    }

  int w=(img->source->AskWidth() +15)/16;
  int h=(img->source->AskHeight()+15)/16;

  img->mean    .Create(w,h);
  img->variance.Create(w,h);

  const Pixel*const* pix = img->source->AskBitmap(Bitmap_Y).AskFrame();

  long long totalmean=0;
  long long totalvar =0;

  for (int y0=0;y0<h;y0++)
    {
      for (int x0=0;x0<w;x0++)
        {
          long long var=0,mean=0;
          
          for (int y=0;y<16;y++)
            for (int x=0;x<16;x++)
              {
                int p = pix[y+y0*16][x+x0*16];
                mean += p;
                var  += p*p;
              }
          
          var = var - mean*mean/256;
          
          img->mean    .Ask(y0,x0) = mean;
          img->variance.Ask(y0,x0) = var;
          
          totalmean += mean;
          totalvar  += var /256;
        }
      //      totalmean /= w;
      //      totalvar  /= w;
    }
  //  totalmean /= h;
  //  totalvar  /= h;

  img->total_mean     = totalmean;
  img->total_variance = totalvar;

#if 0
  double newmean=((double)totalmean)/(w*h*256);
  double newvar =((double)totalvar )/(w*h*256);
  cout << m_framenr << '\t' << newmean << '\t' << newvar;
  if (m_framenr==0)
    cout << "\t0\t0\n";
  else
    cout << "\t" << (newmean<lastmean ? lastmean-newmean : newmean-lastmean)
         << "\t" << (newvar <lastvar  ? lastvar -newvar  : newvar -lastvar ) << endl;
  lastmean=newmean;
  lastvar=newvar;
#endif

  // cerr << m_framenr << " -> " << img->total_mean << endl;
}
