/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "control/taskwrite.hh"
#include "control/datapool.hh"
#include "mpeg/syntax.hh"
#include "mpeg/constants.hh"
#include "input/preproc.hh"
#include "output/streamsink.hh"
#include "coding/ratecontrol.hh"
#include "options.hh"

#include "visualize/writeppm.hh"
#include <fstream.h>

void Task_WriteFrame::Execute(ProcessData& data)
{
  if (options.ProgressDisplay >= Options::ProgressDisplay_Normal)
    {
      ShowTaskInformation("create bitstream",data.threadnr); cerr << "frame " << m_framenr << endl;
      EndShowTaskInformation();
    }

  ImageData* img = ImageData::GetNr(m_framenr);
  assert(img);

  MPEGSyntaxCoder syntaxcoder(img->output);

  SequenceHeaderData sh;
  SeqHdrPool::AskSeqHdr(m_framenr,sh);
  syntaxcoder.SetSequenceHeader(sh);

  // ggf. SeqHdr schreiben

  if (SeqHdrPool::IsSeqHdrBeforeFrame(m_framenr))
    {
      syntaxcoder.WriteSequenceHeader();
    }

  if (img->IsGOPBeginning())
    {
      GOPHeaderData gh;
      img->GetGOPHdr(gh);
      syntaxcoder.WriteGOPHeader(gh);
    }

  syntaxcoder.WritePicture(*img);

  if (options.Enable_CBR_RateControl)
    AdjustFramesizePrediction(img);
}


void Task_WriteToStream::Execute(ProcessData& data)
{
  ImageData* img = ImageData::GetNr(m_framenr);
  assert(img);
  img->output.Flush();

  /*
      uint32 correct=0;
      for (int y=0;y<img->mblks.AskHeight();y++)
        for (int x=0;x<img->mblks.AskWidth();x++)
          {
            Macroblock& mb = img->mblks.Ask(y,x);
            
            correct += BitsToCodeCoeff(mb,img->pichdr);
          }

      cout << "------->>>>>>>>>>>>>> BitCount: " << correct << " bits = " << correct/8 << " Bytes" << endl;
  */

  if (options.ProgressDisplay >= Options::ProgressDisplay_Minimal)
    {
      ShowTaskInformation("write bitstream",data.threadnr);
      cerr << "frame " << m_framenr
           << " ------------------------ (" << img->output.d_bufferidx << " bytes)\n";
      EndShowTaskInformation();
    }

  if (options.Enable_CBR_RateControl)
    AddBufferMismatch(8*(img->desired_size - img->output.d_bufferidx));


  if (0)
    {
      switch (img->pichdr.m_picture_coding_type)
        {
        case PICTYPE_I:
          cout << "framesize " << m_framenr
               << " " << img->output.d_bufferidx
               << " " << img->output.d_bufferidx << " 0 0\n"; break;
        case PICTYPE_P:
          cout << "framesize " << m_framenr
               << " " << img->output.d_bufferidx
               << " 0 " << img->output.d_bufferidx << " 0\n"; break;
        case PICTYPE_B:
          cout << "framesize " << m_framenr
               << " " << img->output.d_bufferidx
               << " 0 0 " << img->output.d_bufferidx << endl; break;
        }

      // Extract-Command:  grep "wr" x | cut -c4- | sort -n >x.dat
      // GNUPLOT-Command:  plot 'x.dat' with impulses, 'x.dat' using 1:3 with impulses, 'x.dat' using 1:4 with impulses, 'x.dat' using 1:4 with impulses
    }

  m_sink.WriteBuffer(img->output.d_buffer,img->output.d_bufferidx);

#if 0
  char buf[100];
  sprintf(buf,"/home/dirk/vis/vis%03d.ppm",m_framenr);
  cout << "Writing VIS\n";
  ofstream of(buf);
  Image<Pixel>* pix = CreateVisualizationImage(*img->source);
  FillMotionVectors(*pix,*img);
  //FillBlockTypes(*pix,*img);
  WriteImageAsPPM(*pix,of);
  delete pix;
#endif
}
