/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/
/*
  Bemerkungen:
  - Die Groesse der real erzeugten Bitmap wird sowohl horizontal als auch vertikal auf ganze
    16er-Werte aufgerundet. AskWidth() und AskHeight() liefern aber die genauen (ggf. kleineren) Werte.
 */

#ifndef IMAGE_BITMAP_HH
#define IMAGE_BITMAP_HH

#include "types.hh"

template <class Pel> class Bitmap
{
public:
  Bitmap()
    {
      d_width = d_height = 0;
      d_rounded_width = d_rounded_height = 0;
      d_frameptr = d_top_field = d_bottom_field = 0;
    }

  Bitmap(int width,int height)
    {
      CreateBitmap(width,height);
    }

  ~Bitmap() { FreeBitmap(); }
  void CreateBitmap(int width,int height);
  void FreeBitmap();      // Bitmap wird entfernt und der Speicher freigegeben.

  void TransferBitmap(Bitmap& bm) // Wird durch andere Bitmap ersetzt, die andere Bitmap wird freigegeben.
    {
      FreeBitmap();
      *this = bm;
      bm.d_frameptr=NULL;
      bm.d_top_field=NULL;
      bm.d_bottom_field=NULL;
    }

  void AlignSize()
    {
      d_width  -= d_width%16;
      d_height -= d_height%16;
    }

        Pel** AskFrame()             { return d_frameptr; }
  const Pel*const * AskFrame_const() const { return d_frameptr; }
        Pel** AskField(bool top)             { return top ? d_top_field : d_bottom_field; }
  const Pel*const * AskField_const(bool top) const { return top ? d_top_field : d_bottom_field; }

  int AskWidth()  const { return d_width;  }
  int AskHeight() const { return d_height; }

  int AskRoundedWidth()  const { return d_rounded_width;  }
  int AskRoundedHeight() const { return d_rounded_height; }

  /* Fuellt den Bereich zwischen der tatsaechlichen und der aufgerundeten Bildgroesse mit einem
     konstanten Wert. */
  void FillBorder(Pel val);

  /* Fuellt den Bereich zwischen der tatsaechlichen und der aufgerundeten Bildgroesse mit dem
     Duplikat der jeweils letzten angenzenden Pixelreihe. */
  void DuplicateBorder();

private:
  int     d_width,d_height;
  int     d_rounded_width, d_rounded_height;
  Pel** d_frameptr;
  Pel** d_top_field;
  Pel** d_bottom_field;
};

#endif
