/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "image/colorspaces.hh"
#include "image/image.hh"
#include "mpeg/constants.hh"

#if USE_MMX
#include "arch/MMX/mmx.hh"
#include "arch/MMX/mmxlib.icc"
#endif

#include <iostream.h>

/*  aus anderem Programm:
 #define Y(r,g,b) ((30*r+59*g+11*b)/100)
 #define U(y,b)   (3*493*(b-y)/4096+128)
 #define V(y,r)   (3*877*(r-y)/4096+128)
*/

/* Matrixcoeffizienten enthalten schon den Skalierungsfaktor 219/255 bzw. 224/255 */

inline void RGB2Y(Pixel r,Pixel g,Pixel b,Pixel &y)
{
  //y  = (Pixel)( 0.2125*r + 0.7154*g + 0.0721*b + 16.5 ); // 1
  y  = (Pixel)( 0.257*r + 0.504*g + 0.098*b + 16.5 ); // 5
}

inline void RGB2CrCb(Pixel r,Pixel g,Pixel b,
		     Pixel& cb,Pixel& cr)
{
  //cb = (Pixel)( 0.500*r - 0.454*g - 0.046*b + 128.5); // 1
  //cr = (Pixel)(-0.115*r - 0.386*g + 0.500*b + 128.5);
  cb = (Pixel)( 0.439*r - 0.368*g - 0.071*b + 128.5); // 5
  cr = (Pixel)(-0.148*r - 0.291*g + 0.439*b + 128.5);

#if 0
  assert(y>=16);
  assert(y<=235);
  assert(cr>=16);
  assert(cr<=240);
  assert(cb>=16);
  assert(cb<=240);
#endif
}

void TransformRGB2YCrCb(Image<Pixel>& img,uint2 chromaformat,bool IsMPEG1)
{
  assert(img.AskColorspace()==RGB_Undefined);

  Bitmap<Pixel>& r = img.AskBitmap(Bitmap_Red);
  Bitmap<Pixel>& g = img.AskBitmap(Bitmap_Green);
  Bitmap<Pixel>& b = img.AskBitmap(Bitmap_Blue);
  Bitmap<Pixel>& yy = img.AskBitmap(Bitmap_Y);
  Bitmap<Pixel>& cr = img.AskBitmap(Bitmap_Cr);
  Bitmap<Pixel>& cb = img.AskBitmap(Bitmap_Cb);

  Pixel** pr = r.AskFrame();
  Pixel** pg = g.AskFrame();
  Pixel** pb = b.AskFrame();
  Pixel** py  = yy.AskFrame();
  Pixel** pcr = cr.AskFrame();
  Pixel** pcb = cb.AskFrame();

  int w = img.AskRoundedWidth();
  int h = img.AskRoundedHeight();

  switch (chromaformat)
    {
    case CHROMA_420:
      if (IsMPEG1)
	{
	  for (int y=0,dy=0;y<h;dy++,y+=2)
	    for (int x=0,dx=0;x<w;dx++,x+=2)
	      {
		Pixel y1,y2,y3,y4;
		RGB2Y(pr[y  ][x]  ,pg[y  ][x  ],pb[y  ][x  ], y1);
		RGB2Y(pr[y  ][x+1],pg[y  ][x+1],pb[y  ][x+1], y2);
		RGB2Y(pr[y+1][x]  ,pg[y+1][x]  ,pb[y+1][x],   y3);
		RGB2Y(pr[y+1][x+1],pg[y+1][x+1],pb[y+1][x+1], y4);
		
		RGB2CrCb((pr[y][x]+pr[y+1][x]+pr[y][x+1]+pr[y+1][x+1])/4,
			 (pg[y][x]+pg[y+1][x]+pg[y][x+1]+pg[y+1][x+1])/4,
			 (pb[y][x]+pb[y+1][x]+pb[y][x+1]+pb[y+1][x+1])/4,
			 pcr[dy][dx],
			 pcb[dy][dx]);
		
		py[y  ][x  ] = y1;
		py[y  ][x+1] = y2;
		py[y+1][x  ] = y3;
		py[y+1][x+1] = y4;
	      }
	}
      else
	{
	  for (int y=0,dy=0;y<h;dy++,y+=2)
	    for (int x=0,dx=0;x<w;dx++,x+=2)
	      {
		Pixel y1,y2,y3,y4;
		RGB2Y(pr[y  ][x]  ,pg[y  ][x  ],pb[y  ][x  ], y1);
		RGB2Y(pr[y  ][x+1],pg[y  ][x+1],pb[y  ][x+1], y2);
		RGB2Y(pr[y+1][x]  ,pg[y+1][x]  ,pb[y+1][x],   y3);
		RGB2Y(pr[y+1][x+1],pg[y+1][x+1],pb[y+1][x+1], y4);
		
		RGB2CrCb((pr[y][x]+pr[y+1][x])/2,
			 (pg[y][x]+pg[y+1][x])/2,
			 (pb[y][x]+pb[y+1][x])/2,
			 pcr[dy][dx],
			 pcb[dy][dx]);
		
		py[y  ][x  ] = y1;
		py[y  ][x+1] = y2;
		py[y+1][x  ] = y3;
		py[y+1][x+1] = y4;
	      }
	}
      break;

    default:
      assert(0);
      break;
    }

  img.SetColorspace(YCrCb);
}


void SubsampleYCrCb(class Image<Pixel>& img,uint2 chromaformat,bool IsMPEG1)
{
  assert(img.AskColorspace()==YCrCb);

  Bitmap<Pixel>& cr = img.AskBitmap(Bitmap_Cr);
  Bitmap<Pixel>& cb = img.AskBitmap(Bitmap_Cb);

  Pixel** pcr = cr.AskFrame();
  Pixel** pcb = cb.AskFrame();

  int w = img.AskRoundedWidth();
  int h = img.AskRoundedHeight();

  switch (chromaformat)
    {
    case CHROMA_420:
      {
        Bitmap<Pixel> newcr; newcr.CreateBitmap(w/2,h/2);
        Bitmap<Pixel> newcb; newcb.CreateBitmap(w/2,h/2);
        Pixel** newpcr = newcr.AskFrame();
        Pixel** newpcb = newcb.AskFrame();

        if (IsMPEG1)
          {
#if USE_MMX
            if (MMX_Supported)
              {
                Prepare_HVMean8_8();
                
                for (int y=0,dy=0;y<h;dy++,y+=2)
                  for (int x=0,dx=0;x<w;dx+=8,x+=16)
                    {
                      HVMean8_8(&pcr[y][x],&pcr[y+1][x],&newpcr[dy][dx]);
                      HVMean8_8(&pcb[y][x],&pcb[y+1][x],&newpcb[dy][dx]);
                    }
                
                EMMS();
              }
            else
#endif
              {
                for (int y=0,dy=0;y<h;dy++,y+=2)
                  for (int x=0,dx=0;x<w;dx++,x+=2)
                    {
                      newpcr[dy][dx] = (pcr[y][x]+pcr[y+1][x]+pcr[y][x+1]+pcr[y+1][x+1])/4;
                      newpcb[dy][dx] = (pcb[y][x]+pcb[y+1][x]+pcb[y][x+1]+pcb[y+1][x+1])/4;
                    }
              }
          }
        else
          {
#if USE_MMX
            if (MMX_Supported)
              {
                Prepare_VMeanHSubsample8_8();
                
                for (int y=0,dy=0;y<h;dy++,y+=2)
                  for (int x=0,dx=0;x<w;dx+=8,x+=16)
                    {
                      VMeanHSubsample8_8(&pcr[y][x],&pcr[y+1][x],&newpcr[dy][dx]);
                      VMeanHSubsample8_8(&pcb[y][x],&pcb[y+1][x],&newpcb[dy][dx]);
                    }
                
                EMMS();
              }
            else
#endif
              {
                for (int y=0,dy=0;y<h;dy++,y+=2)
                  for (int x=0,dx=0;x<w;dx++,x+=2)
                    {
                      newpcr[dy][dx] = (pcr[y][x]+pcr[y+1][x])/2;
                      newpcb[dy][dx] = (pcb[y][x]+pcb[y+1][x])/2;
                    }
              }
          }

        cr.TransferBitmap(newcr);
        cb.TransferBitmap(newcb);

        img.m_subsampledH=true;
        img.m_subsampledV=true;
      }
      break;

    default:
      assert(0);
      break;
    }
}
