/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "input/loadimg_ppm.hh"
#include "image/image.hh"
#include <iostream.h>
#include <stdlib.h>
#include <string.h>

const int MaxLineLen=100;

void ImageLoader_PPM::Load(istream& istr,Image<Pixel>& img)
{
  char buf[MaxLineLen];

  istr.getline(buf,MaxLineLen);
  if (buf[0] != 'P' || buf[1] != '6')
    throw "Invalid image file (not a 'P6' PPM, 3 bytes per color, rawbits - file).";

  for (;;) { istr.getline(buf,MaxLineLen); if (buf[0] != '#') break; }

  int width  = atoi(strtok(buf ," \t"));
  int height = atoi(strtok(NULL," \t"));

  for (;;) { istr.getline(buf,MaxLineLen); if (buf[0] != '#') break; }

  int maxval = atoi(strtok(buf," \t"));
  
  if (maxval != 255)
    throw "Invalid image value range (0-255 expected).";

  img.CreateImage(RGB_Undefined , width,height);


  unsigned char* linebuf = new unsigned char [width*3];


  Bitmap<Pixel>& r = img.AskBitmap(Bitmap_Red);
  Bitmap<Pixel>& g = img.AskBitmap(Bitmap_Green);
  Bitmap<Pixel>& b = img.AskBitmap(Bitmap_Blue);

  for (int y=0;y<height;y++)
    {
      Pixel* rp = r.AskFrame()[y];
      Pixel* gp = g.AskFrame()[y];
      Pixel* bp = b.AskFrame()[y];

      istr.read(linebuf,3*width);

      for (int x=0;x<width;x++)
	{
	  rp[x] = linebuf[3*x  ];
	  gp[x] = linebuf[3*x+1];
	  bp[x] = linebuf[3*x+2];
	}
    }
  
  delete[] linebuf;
}
