/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "input/loadimg_tga.hh"
#include "image/image.hh"
#include <iostream.h>
#include <stdlib.h>
#include <string.h>

void ImageLoader_TGA::Load(istream& istr,Image<Pixel>& img)
{
   int  ImageIDLength = istr.get();
  uint8 ColorMapType  = istr.get();
  uint8 ImageType     = istr.get();

  switch (ImageType)
    {
    case 2:  // uncompressed TrueColor
      break;

    case 0:
    case 1:  // uncompressed CLUT
    case 3:  // uncompressed BW
    case 9:  // rle CLUT
    case 10: // rle TrueColor
    case 11: // rle BW
      throw "TGA file type not supported.";

    default:
      throw "No Targa format input file.";
    }

  if (ColorMapType!=0)   // Dieser Test gilt nur, wenn ausschliesslich TrueColor-Bilder erlaubt sind.
    throw "No Targa format input file.";

  // CLUT-Info ueberlesen
  {for (int i=0;i<5;i++)
    istr.get();}

  // Origins ueberlesen
  {for (int i=0;i<4;i++)
    istr.get();}

  int Width  = istr.get(); Width  |= istr.get()<<8;
  int Height = istr.get(); Height |= istr.get()<<8;

  int PixelDepth = istr.get();
  if (PixelDepth != 24 && PixelDepth != 32)
    throw "Targa image pixel depth not supported.";
  int PixelBytes = PixelDepth/8;

  int imgdescr   = istr.get(); // ImageDescriptor

  if (imgdescr & 0x10 != 0)
    throw "Targa image origin not supported.";

  // ImageID ueberlesen
  {for (int i=0;i<ImageIDLength;i++)
    istr.get();}

  img.CreateImage(RGB_Undefined , Width,Height);


  unsigned char* linebuf = new unsigned char [Width*PixelBytes];


  Bitmap<Pixel>& r = img.AskBitmap(Bitmap_Red);
  Bitmap<Pixel>& g = img.AskBitmap(Bitmap_Green);
  Bitmap<Pixel>& b = img.AskBitmap(Bitmap_Blue);

  for (int y=0;y<Height;y++)
    {
      int yy;

      if ((imgdescr & 0x30) == 0x00) { yy=Height-1-y; }
      else                           { yy=y; }

      Pixel* rp = r.AskFrame()[yy];
      Pixel* gp = g.AskFrame()[yy];
      Pixel* bp = b.AskFrame()[yy];

      istr.read(linebuf,Width*PixelBytes);

      for (int x=0;x<Width*PixelBytes;x+=PixelBytes)
	{
	  *bp++ = linebuf[x  ];
	  *gp++ = linebuf[x+1];
	  *rp++ = linebuf[x+2];
	}
    }
  
  delete[] linebuf;
}
