/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#ifndef DISABLE_MJPEG
#define DISABLE_MJPEG 0
#endif

#if !DISABLE_MJPEG

#define AUDIO 1

#include "input/preproc_avi.hh"
#include "image/image.hh"
#include "options.hh"
#include <iostream.h>
#include <stdlib.h>
#include <string.h>

#include "input/avi_tables.icc"
void jpeg_short_Buffer_src (j_decompress_ptr cinfo, InputPreprocessor_AVI* preproc);


uint32 InputPreprocessor_AVI::ReadID()
{
  uint32 id;
  id  = ((uint32)infile.get())<<24;
  id |= ((uint32)infile.get())<<16;
  id |= ((uint32)infile.get())<< 8;
  id |= ((uint32)infile.get());

  return id;
}

uint32 InputPreprocessor_AVI::ReadDWORD()
{
  uint32 id;
  id  = ((uint32)infile.get());
  id |= ((uint32)infile.get())<< 8;
  id |= ((uint32)infile.get())<<16;
  id |= ((uint32)infile.get())<<24;

  return id;
}

inline uint64 Align2(uint64 n)
{
  if (n&1) return n+1; else return n;
}




InputPreprocessor_AVI::InputPreprocessor_AVI()
  : d_FirstFrame(0),
    d_LastFrame(MASK31),
    d_smallmode(false),
    d_EndOfFileReached(false)
{
  d_lastframe.buf  = NULL;
  d_lastframe.size = 0;

#if AUDIO
  if (options.mjpeg_avi.SaveAudio)
    audiostream.open("audio.pcm");
#endif
}

InputPreprocessor_AVI::~InputPreprocessor_AVI()
{
  jpeg_destroy_decompress(&cinfo);
  infile.close();

  if (d_lastframe.buf)
    delete[] d_lastframe.buf;

#if AUDIO
  if (options.mjpeg_avi.SaveAudio)
    audiostream.close();
#endif
}

void InputPreprocessor_AVI::ReadChunk()
{
  uint32 id1    = ReadID();
  uint32 length = ReadDWORD();

  struct Pos pos;
  pos.current_chunk = id1;

  // RIFF und LIST-Chunks haben Untertypen, die direkt nach der Laenge angegeben werden.

  if (id1 == ID_RIFF || id1 == ID_LIST)
    {
      pos.current_chunk_type = ReadID();
      length-=4;
    }
  else
    pos.current_chunk_type = 0;

  pos.chunk_length  = length;
  pos.abs_chunk_end = infile.tellg()+Align2(length);
  d_chkqueue.AppendAtHead(pos);

  // cout << ((void*)pos.current_chunk) << " " << ((void*)pos.current_chunk_type) << endl;

  if (pos.current_chunk == ID_JUNK)
    {
      EndChunk();
      ReadChunk();
    }
}

void InputPreprocessor_AVI::EndChunk()
{
  infile.seekg(d_chkqueue.AskHead().abs_chunk_end);
  d_chkqueue.RemoveHead();
}

bool InputPreprocessor_AVI::AreAtEndOfChunk() const
{
  return (Align2(const_cast<ifstream&>(infile).tellg()) == d_chkqueue.AskHead().abs_chunk_end);
}

void InputPreprocessor_AVI::ReadAVIHeader(bool firstheader)
{
  ReadChunk();

  if (d_chkqueue.AskHead().current_chunk      != ID_RIFF ||
      d_chkqueue.AskHead().current_chunk_type != ID_AVI)
    {
      cerr << "Not an AVI input file\n";
      exit(10);
    }

  d_chkqueue.AskHead().state = AVI_BeforeHDRL;
}

void InputPreprocessor_AVI::ReadHDRL()
{
  // ---- HDRL-List-Header laden ----

  ReadChunk();

  if (d_chkqueue.AskHead().current_chunk      != ID_LIST ||
      d_chkqueue.AskHead().current_chunk_type != ID_hdrl)
    { cerr << "AVI file structure error: hdrl-LIST-chunk missing.\n"; exit(10); }



  // ---- avih-Chunk laden ----

  ReadChunk();
  if (d_chkqueue.AskHead().current_chunk != ID_avih)
    { cerr << "AVI file structure error: avih-chunk missing.\n"; exit(10); }

  uint32 TimeBtwFrames = ReadDWORD();
  d_framerate100 = 100000000 / TimeBtwFrames;
  cerr << "Framerate: " << d_framerate100/100 << "," << d_framerate100%100 << " fps\n";
  ReadDWORD(); // MaximumDataRate
  d_padding    = ReadDWORD();
  d_avih_flags = ReadDWORD();

  if (d_avih_flags & AVIH_HAS_IDX1)    cerr << "AVI has idx1-chunk\n";
  if (d_avih_flags & AVIH_USE_INDEX)   cerr << "AVI is not linear\n";
  if (d_avih_flags & AVIH_INTERLEAVED) cerr << "AVI is interleaved\n";
  if (d_avih_flags & AVIH_CAPTURE)     cerr << "AVI is optimized for live-capture\n";
  if (d_avih_flags & AVIH_COPYRIGHTED) cerr << "AVI file contents is copyrighted\n";

  d_nPicsInStream = ReadDWORD();
  cerr << "AVI contains " << d_nPicsInStream << " frames\n";

  ReadDWORD(); // Initial Frames

  d_nStreams = ReadDWORD();
  cerr << "AVI contains " << d_nStreams << " streams\n";
  ReadDWORD(); // suggested Buffer Size
  
  d_width  = ReadDWORD();
  d_height = ReadDWORD();
  cerr << "AVI size is " << d_width << 'x' << d_height << endl;

  d_IsInterlaced = (d_height>350);

  uint64 timescale = ReadDWORD(); // TimeScale
  cerr << "TimeScale: " << timescale << endl;
  uint64 datarate  = ReadDWORD(); // DataRate
  cerr << "DataRate:  " << datarate << endl;
  ReadDWORD(); // StartTime
  uint64 datalength = ReadDWORD(); // DataLength
  cerr << "DataLen:   " << datalength << endl;

  EndChunk();


  // ----- strl-LIST lesen -----

  d_VideoStreamIdx=-1;

  for (uint32 i=0;i<d_nStreams;i++)
    {
      ReadChunk();
      if (d_chkqueue.AskHead().current_chunk      != ID_LIST ||
          d_chkqueue.AskHead().current_chunk_type != ID_strl)
        { cerr << "AVI file structure error: strl-LIST-chunk missing.\n"; exit(10); }

      ReadChunk();
      if (d_chkqueue.AskHead().current_chunk != ID_strh)
        { cerr << "AVI file structure error: strh-chunk missing.\n"; exit(10); }

      d_streamtype[i] = ReadID();
      switch (d_streamtype[i])
        {
        case ID_vids:
          cerr << "AVI stream " << i << " is a video stream\n";
          if (d_VideoStreamIdx<0)
            { d_VideoStreamIdx = i; }   // Dieser Videostream wird geladen.
          break;
        case ID_auds: cerr << "AVI stream " << i << " is a audio stream\n"; break;
        default:      cerr << "AVI stream " << i << " is a stream of unknown type\n"; break;
        }

      d_codec[i] = ReadID();
      if (d_streamtype[i]==ID_vids)
        {
          cerr << "AVI stream coding is: ";
          switch(d_codec[i])
            {
            case ID_MJPG: cerr << "MotionJPEG\n"; break;
            case ID_0000: cerr << "uncompressed\n"; break;
            default:      cerr << "unknown (" << ((void*)d_codec[i]) << ")\n"; break;
            }
        }

      /*uint64 Flags =*/ ReadDWORD();
      /*uint64 Priority =*/ ReadDWORD();
      /*uint64 InitialFrames =*/ ReadDWORD();
      uint64 TimeScale = ReadDWORD();  cerr << "TimeScale " << TimeScale << endl;
      uint64 DataRate = ReadDWORD();   cerr << "DataRate " << DataRate << endl;
      /*uint64 StartTime =*/ ReadDWORD();
      uint64 DataLength = ReadDWORD(); cerr << "DataLength " << DataLength << endl;
      ReadDWORD();
      ReadDWORD();
      uint64 SampleSize = ReadDWORD(); cerr << "SampleSize " << SampleSize << endl;

      EndChunk();

      ReadChunk();
      if (d_chkqueue.AskHead().current_chunk != ID_strf)
        { cerr << "AVI file structure error: strf-chunk missing.\n"; exit(10); }
      EndChunk();

      EndChunk(); // strl
    }

  EndChunk(); // hdrl

  assert(d_chkqueue.AskHead().current_chunk == ID_RIFF);
}

void InputPreprocessor_AVI::ProcessFileUntilData(bool firsttime)
{
  while (d_chkqueue.IsEmpty() || d_chkqueue.AskHead().state != BeforeData)
    {
      if (infile.eof())
        {
          d_EndOfFileReached=true;
          return;
        }

      if (d_chkqueue.IsEmpty())
        ReadAVIHeader(firsttime);
      else
        {
          Pos p = d_chkqueue.AskHead();

          if (p.current_chunk==ID_RIFF && p.state==AVI_BeforeHDRL)
            {
              ReadHDRL();
              ReadChunk();
              if (d_chkqueue.AskHead().current_chunk      != ID_LIST ||
                  d_chkqueue.AskHead().current_chunk_type != ID_movi)
                { cerr << "AVI file structure error: movi-LIST-chunk missing.\n"; exit(10); }

              ReadChunk();
              if (d_chkqueue.AskHead().current_chunk == ID_LIST)
                {
                  assert(d_chkqueue.AskHead().current_chunk_type==ID_rec);
                  ReadChunk();
                }

              d_chkqueue.AskHead().state = BeforeData;
            }
          else
            { assert(0); }
        }
    }


  uint32 chunkid = d_chkqueue.AskHead().current_chunk;

  switch (chunkid & 0xFFFF)
    {
    case (ID___db & 0xFFFF): cerr << "Video-Data-Chunk "
                                  << ((char)((chunkid>>24)&0xFF))
                                  << ((char)((chunkid>>24)&0xFF)) << " db\n";
    break;
    case (ID___dc & 0xFFFF): cerr << "Video-Data-Chunk "
                                  << ((char)((chunkid>>24)&0xFF))
                                  << ((char)((chunkid>>24)&0xFF)) << " dc\n";
    break;
    case (ID___wb & 0xFFFF): cerr << "Audio-Data-Chunk "
                                  << ((char)((chunkid>>24)&0xFF))
                                  << ((char)((chunkid>>24)&0xFF)) << " wb\n";
    break;
    default:
      assert(0);
      break;
    }
}

InputPreprocessor_AVI::      // <<--- Fuer SUN auskommentieren
Buffer InputPreprocessor_AVI::GetNextBuffer()
{
  if (!d_bufqueue.IsEmpty())
    {
      Buffer buf = d_bufqueue.AskHead();
      d_bufqueue.RemoveHead();
      return buf;
    }

  if (infile.eof())
    {
      Buffer buf;
      buf.size=0;
      buf.buf=0;
      return buf;
    }

  uint32 chunkid = d_chkqueue.AskHead().current_chunk;
  switch (chunkid & 0xFFFF)
    {
    case (ID___db & 0xFFFF):
    case (ID___dc & 0xFFFF):
      {
        Buffer buf;
        buf.size = d_chkqueue.AskHead().chunk_length;

        if (buf.size>0)
          {
            buf.buf  = new uint8[buf.size];
            infile.read(buf.buf , buf.size);
            EndChunk();

            if (d_codec[d_VideoStreamIdx]==ID_MJPG)
              {
                int i;
                for (i=0;buf.buf[i] != 0xFF || buf.buf[i+1] != 0xC0 ;i++)
                  {
                  }
                
                int height = (((int)buf.buf[i+5])<<8)|buf.buf[i+6];
                
                d_IsInterlaced = (height<d_height);
                
                if (height > ((d_height+15)/16)*16)
                  {
                    // Miro-NT-Bug umgehen
                    
                    height /= 2;
                    
                    buf.buf[i+5] = (height>>8);
                    buf.buf[i+6] = (height&0xFF);
                  }
                
                while (buf.buf[buf.size-1]==0xFF)
                  buf.size--;
              }

            if (d_lastframe.buf)
              delete[] d_lastframe.buf;
            d_lastframe = buf;
            
            ReadChunk();
            
            return buf;
          }
        else
          {
            EndChunk();
            ReadChunk();

            return d_lastframe;
          }
      }
    break;

    case (ID___wb & 0xFFFF):
      if (options.mjpeg_avi.SaveAudio)
        {
          if (options.ProgressDisplay >= Options::ProgressDisplay_Maximal)
            cerr << "NOTE: saving audio data\n";

          int size = d_chkqueue.AskHead().chunk_length;
          uint8* audiobuffer = new uint8[size];
          infile.read(audiobuffer , size);

#if AUDIO
          for (int i=0;i<size;i+=2)
            {
              audiostream.put(audiobuffer[i+1]);
              audiostream.put(audiobuffer[i  ]);
            }
#endif
        }
      else
        {
          if (options.ProgressDisplay >= Options::ProgressDisplay_Maximal)
            cerr << "NOTE: skipping audio chunk\n";
        }
        
      EndChunk();
      ReadChunk();
      return GetNextBuffer();
    break;

    default:
      {
        uint8 id[5];
        id[0] = (chunkid>>24)&0xFF;
        id[1] = (chunkid>>16)&0xFF;
        id[2] = (chunkid>> 8)&0xFF;
        id[3] = (chunkid    )&0xFF;
        id[4] = 0;

        if (options.ProgressDisplay >= Options::ProgressDisplay_Normal)
          cerr << "NOTE: skipping unknown chunk: " << ((void*) chunkid) << " " << id << endl;
      }

      EndChunk();
      ReadChunk();
      return GetNextBuffer();
    }

  assert(0);
  Buffer buf;
  return buf;
}

void InputPreprocessor_AVI::SetInputFile(const char* filename,bool interlaced)
{
  infile.open(filename);
  if (!infile) {
    fprintf(stderr, "can't open %s\n", filename);
    exit(0);
  }


  ProcessFileUntilData(true);
  cerr << d_nPicsInStream << " Bilder im AVI-Stream\n";


  // Codec spezifische Initialisierungen

  if (d_VideoStreamIdx >=0 && d_codec[d_VideoStreamIdx]==ID_MJPG)
    {
      cerr << "Initialize MJPG codec.\n";

      Buffer huffbuf;
      huffbuf.buf = new uint8[hufftbl_length];
      
      memcpy(huffbuf.buf,hufftbl,hufftbl_length);
      huffbuf.size = hufftbl_length;
      d_bufqueue.Append(huffbuf);

      cinfo.err = jpeg_std_error(&jerr);

      jpeg_create_decompress(&cinfo);
      jpeg_short_Buffer_src(&cinfo, this);
      (void) jpeg_read_header(&cinfo, FALSE);
    }

  piccnt=0;
}


void InputPreprocessor_AVI::SetFrameRange(uint32 firstframe,uint32 lastframe)
{
  d_FirstFrame = firstframe;
  d_LastFrame  = lastframe;
}

void InputPreprocessor_AVI::SetSmallMode()
{
#if 0
  d_smallmode = true;
  nPicsInStream *= 2;     // Kollidiert mit 'if(interlaced)' oben, das aber noch nicht benutzt wird.

  if (d_FirstFrame&1 != d_LastFrame&1)
    d_LastFrame--;
#endif
}

void InputPreprocessor_AVI::Init()
{
#if 0
  d_nextimg = GetNextImage(NULL);
  d_width  = d_nextimg->AskWidth();
  d_height = d_nextimg->AskHeight();

  cout << "WIDTH/HEIGHT: " << d_width << '/' << d_height << endl;
#endif
}


bool         InputPreprocessor_AVI::IsProgressiveSequence() const
{
  return true;
}

int          InputPreprocessor_AVI::AskNextImageNum() const { return piccnt; }
const char*  InputPreprocessor_AVI::AskNextImageFilename() const { return "UNAVAILABLE"; }

bool         InputPreprocessor_AVI::HasMoreInput() const
{
#if 1
  // cout << "HASMOREINPUT: " << piccnt << " " << nPicsInStream << endl;
  return piccnt < min(d_LastFrame,d_nPicsInStream) && !((ifstream&)(infile)).eof();
#endif
}


class Image<Pixel>* InputPreprocessor_AVI::ReadImage_MJPG(class Image<Pixel>* img,bool SkipFrame)
{
  if (img==NULL)
    {
      img = new Image<Pixel>;
      img->CreateImage(YCrCb,
                       d_width  / (d_smallmode ? 2 : 1),
                       d_height / (d_smallmode ? 2 : 1));
                       /*
                       cinfo.output_width            / (d_smallmode ? 2 : 1),
                       cinfo.output_height * nFields / (d_smallmode ? 2 : 1));
                       */
    }



  int row_stride;		/* physical row width in output buffer */
  JSAMPARRAY buffer;		/* Output row buffer */

  int nFields = (d_IsInterlaced ? 2 : 1);

  for (int field=0;field<nFields;field++)
    {
      cout << "Y0\n";

      (void) jpeg_read_header(&cinfo, FALSE);
      /* Bild ist ja schon im YCbCr-Farbraum, also nicht umwandeln. */
      cinfo.out_color_space = JCS_YCbCr;

      cout << "Y1\n";
      
      (void) jpeg_start_decompress(&cinfo);
      cout << "Y2\n";
      
      row_stride = cinfo.output_width * cinfo.output_components;
      
      buffer = (*cinfo.mem->alloc_sarray)
        ((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);
      
      Pixel** y=0;
      Pixel** cb=0;
      Pixel** cr=0;
      
      if (!SkipFrame)
        {
          if (!d_IsInterlaced)
            {
              y  = img->AskBitmap(Bitmap_Y ).AskFrame();
              cb = img->AskBitmap(Bitmap_Cb).AskFrame();
              cr = img->AskBitmap(Bitmap_Cr).AskFrame();
            }
          else
            {
              y  = img->AskBitmap(Bitmap_Y ).AskField(field==0);
              cb = img->AskBitmap(Bitmap_Cb).AskField(field==0);
              cr = img->AskBitmap(Bitmap_Cr).AskField(field==0);
            }
        }
      
      cout << "Y3\n";

      while (cinfo.output_scanline < cinfo.output_height) {
        (void) jpeg_read_scanlines(&cinfo, buffer, 1);
        
        if (!SkipFrame)
          {
            int xx=0;
            Pixel* py  = y [cinfo.output_scanline-1]; 
            Pixel* pcb = cb[cinfo.output_scanline-1]; 
            Pixel* pcr = cr[cinfo.output_scanline-1]; 
            
            if (d_smallmode)
              for (unsigned int x=0;x<cinfo.output_width;x++)
                {
                  *py++  = buffer[0][xx++];
                  *pcb++ = buffer[0][xx++];
                  *pcr++ = buffer[0][xx++];
                  
                  xx+=3; x++;
                }
            else
              {
                JOCTET* bufp = buffer[0];
                
                for (unsigned int x=0;x<cinfo.output_width;x++)
                  {
                    *py++  = *bufp++;
                    *pcb++ = *bufp++;
                    *pcr++ = *bufp++;
                  }
              }
          }
      }
  
      (void) jpeg_finish_decompress(&cinfo);
    }

  return img;
}

class Image<Pixel>* InputPreprocessor_AVI::ReadImage_RAW (class Image<Pixel>* img,bool SkipFrame)
{
  if (img==NULL)
    {
      img = new Image<Pixel>;
      img->CreateImage(RGB_Undefined,
                       d_width  / (d_smallmode ? 2 : 1),
                       d_height / (d_smallmode ? 2 : 1));
    }


  //int nFields = (d_IsInterlaced ? 2 : 1);

  //for (int field=0;field<nFields;field++)
    {
      InputPreprocessor_AVI::Buffer buf = GetNextBuffer();
      uint32 nPixels = d_width*d_height;
      enum { Illegal, AVI15, AVI24 } aviformat = Illegal;

           if (buf.size == nPixels*2) aviformat = AVI15;
      else if (buf.size == nPixels*3) aviformat = AVI24;
      else { delete img; return NULL; }

      if (!SkipFrame)
        {
          Pixel** r=0;
          Pixel** g=0;
          Pixel** b=0;
      
          // if (!d_IsInterlaced)
            {
              r = img->AskBitmap(Bitmap_Red  ).AskFrame();
              g = img->AskBitmap(Bitmap_Green).AskFrame();
              b = img->AskBitmap(Bitmap_Blue ).AskFrame();
            }
          /*
          else
            {
              r = img->AskBitmap(Bitmap_Red  ).AskField(field==0);
              g = img->AskBitmap(Bitmap_Green).AskField(field==0);
              b = img->AskBitmap(Bitmap_Blue ).AskField(field==0);
            }
          */

          if (aviformat == AVI15)
            {
              uint8* bf = buf.buf;
              for (int y=d_height-1;y>=0;y--)
                for (int x=0;x<d_width;x++)
                  {
                    uint16 rgb = (bf[1]<<8)|(bf[0]);
                    bf+=2;

                    r[y][x] = ((rgb & 0x7C00) >> 10) << 3;
                    g[y][x] = ((rgb & 0x03E0) >>  5) << 3;
                    b[y][x] = ((rgb & 0x001F) >>  0) << 3;
                  }
            }
          else
            {
              uint8* bf = buf.buf;
              for (int y=d_height-1;y>=0;y--)
                {
                  for (int x=0;x<d_width;x++)
                    {
                      r[y][x] = bf[0];
                      g[y][x] = bf[1];
                      b[y][x] = bf[2];
                      
                      bf+=3;
                    }
                }
            }
        }
    }

  return img;
}

class Image<Pixel>* InputPreprocessor_AVI::GetNextImage(class Image<Pixel>* reuse)
{
  Image<Pixel>* img = NULL;

  bool SkipFrame=false;

  do
    {
      SkipFrame = (piccnt < d_FirstFrame);

      if (d_smallmode && ((piccnt-d_FirstFrame)&1))
	SkipFrame = true;

      if (!SkipFrame)
        {
          if (reuse)
            {
              assert(0);
              img = reuse;
            }
          else
            {
              img = NULL;
            }
        }
      else
        {
          cerr << "skipping frame " << piccnt << endl;
        }
      
      switch (d_codec[d_VideoStreamIdx])
        {
        case ID_MJPG:
	  img = ReadImage_MJPG(img,SkipFrame);
	  break;
        case ID_0000: img = ReadImage_RAW (img,SkipFrame); break;
        }

      piccnt++;
    } while (piccnt<=d_FirstFrame || SkipFrame);


  // Randbereich ausfuellen

  img->DuplicateBorder();
  img->AlignSize();  // TODO: Quick-Fix

  return img;
}









/************************************* MJPEG - Codec ******************************/



/* Expanded data source object for stdio input */

typedef struct SourceMgr {
  struct jpeg_source_mgr pub;	/* public fields */

  InputPreprocessor_AVI* avireader;
  JOCTET * buffer;		/* start of buffer */
  boolean  start_of_file;
} my_source_mgr;

typedef my_source_mgr * my_src_ptr;

#define INPUT_BUF_SIZE  510*1024


/*
 * Initialize source --- called by jpeg_read_header
 * before any data is actually read.
 */

static void
init_source (j_decompress_ptr cinfo)
{
  my_src_ptr src = (my_src_ptr) cinfo->src;

  /* We reset the empty-input-file flag for each image,
   * but we don't clear the input buffer.
   * This is correct behavior for reading a series of images from one source.
   */
  src->start_of_file = TRUE;
}


/*
 * Fill the input buffer --- called whenever buffer is emptied.
 *
 * In typical applications, this should read fresh data into the buffer
 * (ignoring the current state of next_input_byte & bytes_in_buffer),
 * reset the pointer & count to the start of the buffer, and return TRUE
 * indicating that the buffer has been reloaded.  It is not necessary to
 * fill the buffer entirely, only to obtain at least one more byte.
 *
 * There is no such thing as an EOF return.  If the end of the file has been
 * reached, the routine has a choice of ERREXIT() or inserting fake data into
 * the buffer.  In most cases, generating a warning message and inserting a
 * fake EOI marker is the best course of action --- this will allow the
 * decompressor to output however much of the image is there.  However,
 * the resulting error message is misleading if the real problem is an empty
 * input file, so we handle that case specially.
 *
 * In applications that need to be able to suspend compression due to input
 * not being available yet, a FALSE return indicates that no more data can be
 * obtained right now, but more may be forthcoming later.  In this situation,
 * the decompressor will return to its caller (with an indication of the
 * number of scanlines it has read, if any).  The application should resume
 * decompression after it has loaded more data into the input buffer.  Note
 * that there are substantial restrictions on the use of suspension --- see
 * the documentation.
 *
 * When suspending, the decompressor will back up to a convenient restart point
 * (typically the start of the current MCU). next_input_byte & bytes_in_buffer
 * indicate where the restart point will be if the current call returns FALSE.
 * Data beyond this point must be rescanned after resumption, so move it to
 * the front of the buffer rather than discarding it.
 */

boolean fill_input_buffer (j_decompress_ptr cinfo)
{
  my_src_ptr src = (my_src_ptr) cinfo->src;

  /*
  cout << "PTR: " << ((void*)src->pub.next_input_byte)
       << " BASE: " << ((void*)src->buffer)
       << "   BYTES " << src->pub.bytes_in_buffer << endl;
  */

  InputPreprocessor_AVI::Buffer buf = src->avireader->GetNextBuffer();
  if (buf.size==0)
    return FALSE;

  assert(buf.size <= INPUT_BUF_SIZE); // Buffer muss komplettes Bild aufnehmen koennen.

  assert(sizeof(JOCTET)==1);

  memcpy(src->buffer , buf.buf , buf.size);

  src->pub.next_input_byte = src->buffer;
  src->pub.bytes_in_buffer = buf.size;

  /*
  cout << "SHOW\n";
  for (int i=0;i<10;i++)
    cout << "D " << ((void*)src->buffer[i]) << endl;
  cout << "...\n";
  for (int i=-10;i<0;i++)
    cout << "D " << ((void*)src->buffer[buf.size+i]) << endl;
  */

  //  delete[] buf.buf;

  src->start_of_file=FALSE;
  return TRUE;
}


/*
 * Skip data --- used to skip over a potentially large amount of
 * uninteresting data (such as an APPn marker).
 *
 * Writers of suspendable-input applications must note that skip_input_data
 * is not granted the right to give a suspension return.  If the skip extends
 * beyond the data currently in the buffer, the buffer can be marked empty so
 * that the next read will cause a fill_input_buffer call that can suspend.
 * Arranging for additional bytes to be discarded before reloading the input
 * buffer is the application writer's problem.
 */

static void
skip_input_data (j_decompress_ptr cinfo, long num_bytes)
{
  my_src_ptr src = (my_src_ptr) cinfo->src;

  /* Just a dumb implementation for now.  Could use fseek() except
   * it doesn't work on pipes.  Not clear that being smart is worth
   * any trouble anyway --- large skips are infrequent.
   */
  if (num_bytes > 0) {
    while (num_bytes > (long) src->pub.bytes_in_buffer) {
      num_bytes -= (long) src->pub.bytes_in_buffer;
      (void) fill_input_buffer(cinfo);
      /* note we assume that fill_input_buffer will never return FALSE,
       * so suspension need not be handled.
       */
    }
    src->pub.next_input_byte += (size_t) num_bytes;
    src->pub.bytes_in_buffer -= (size_t) num_bytes;
  }
}


/*
 * An additional method that can be provided by data source modules is the
 * resync_to_restart method for error recovery in the presence of RST markers.
 * For the moment, this source module just uses the default resync method
 * provided by the JPEG library.  That method assumes that no backtracking
 * is possible.
 */


/*
 * Terminate source --- called by jpeg_finish_decompress
 * after all data has been read.  Often a no-op.
 *
 * NB: *not* called by jpeg_abort or jpeg_destroy; surrounding
 * application must deal with any cleanup that should happen even
 * for error exit.
 */

static void
term_source (j_decompress_ptr)
{
  /* no work necessary here */
}


/*
 * Prepare for input from a stdio stream.
 * The caller must have already opened the stream, and is responsible
 * for closing it after finishing decompression.
 */

void jpeg_short_Buffer_src (j_decompress_ptr cinfo, InputPreprocessor_AVI* preproc)
{
  my_src_ptr src;

  /* The source object and input buffer are made permanent so that a series
   * of JPEG images can be read from the same file by calling jpeg_stdio_src
   * only before the first one.  (If we discarded the buffer at the end of
   * one image, we'd likely lose the start of the next one.)
   * This makes it unsafe to use this manager and a different source
   * manager serially with the same JPEG object.  Caveat programmer.
   */
  if (cinfo->src == NULL) {	/* first time for this JPEG object? */
    cinfo->src = (struct jpeg_source_mgr *)
      (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_PERMANENT,
				  sizeof(my_source_mgr));
    src = (my_src_ptr) cinfo->src;
    src->buffer = (JOCTET *)
      (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_PERMANENT,
				  (INPUT_BUF_SIZE+2) * sizeof(JOCTET));
  }

  src = (my_src_ptr) cinfo->src;
  src->pub.init_source = init_source;
  src->pub.fill_input_buffer = fill_input_buffer;
  src->pub.skip_input_data = skip_input_data;
  src->pub.resync_to_restart = jpeg_resync_to_restart; /* use default method */
  src->pub.term_source = term_source;
  src->pub.bytes_in_buffer = 0; /* forces fill_input_buffer on first read */
  src->pub.next_input_byte = NULL; /* until buffer loaded */

  src->avireader = preproc;
}







/* -------------- Das hier fuer SUNs CC --------*/
//template class Queue</*InputPreprocessor_AVI::*/Pos>;
//template class Queue</*InputPreprocessor_AVI::*/Buffer>;

/* ---------------- Das hier fuer GCC ----------*/
template class Queue<InputPreprocessor_AVI::Pos>;
template class Queue<InputPreprocessor_AVI::Buffer>;

#include <utility/queue.cc>

#endif
