/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "input/preproc_sglfrms.hh"
#include "input/loadimg_ppm.hh"
#include "input/loadimg_tga.hh"
#include "image/image.hh"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <iostream.h>
#include <iomanip.h>
#include <fstream.h>


InputPreprocessor_SingleFrames::Parameter::Parameter()
  : m_InputType(ProgressiveFrames),
    m_InputFormat(Format_PPM),
    m_TopFieldFirst(true),
    m_first_num(1),
    m_last_num(9999999),
    m_framerate100(2500)
{
}

InputPreprocessor_SingleFrames::InputPreprocessor_SingleFrames(const Parameter& p,
							       const char* nametemplate)
  : d_parameter(p),
    d_next_num(p.m_first_num),
    d_filename_template(nametemplate),
    d_nextimg(NULL)
{
  GenerateFilename();
  d_nextimg = GetNextImage(NULL);
  d_width  = d_nextimg->AskWidth();
  d_height = d_nextimg->AskHeight();
}


bool InputPreprocessor_SingleFrames::IsProgressiveSequence() const
{
  return d_parameter.m_InputType == InputPreprocessor_SingleFrames::Parameter::ProgressiveFrames;
}


bool InputPreprocessor_SingleFrames::HasMoreInput() const
{
  if (d_next_num > d_parameter.m_last_num)
    return false;

  int fd = open(d_next_file,O_RDONLY);
  if (fd==-1)
    return false;

  close(fd);
  return true;
}

class Image<Pixel>* InputPreprocessor_SingleFrames::GetNextImage(class Image<Pixel>* reuse)
{
  if (d_nextimg)
    {
      if (reuse)
        delete reuse;

      Image<Pixel>* img = d_nextimg;
      d_nextimg=NULL;
      return img;
    }


  ifstream istr(d_next_file);
  if (!istr)
    throw "Could not open image file.";

  Image<Pixel>* img;
  if (reuse)
    {
      img = reuse;
    }
  else
    img = new Image<Pixel>;

  if (d_parameter.m_InputFormat == Parameter::Format_PPM)
    {
      ImageLoader_PPM loader_ppm;
      loader_ppm.Load(istr,*img);
    }
  else if (d_parameter.m_InputFormat == Parameter::Format_TGA)
    {
      ImageLoader_TGA loader_tga;
      loader_tga.Load(istr,*img);
    }
  else
    {
      assert(0);
    }

  switch (d_parameter.m_InputType)
    {
    case InputPreprocessor_SingleFrames::Parameter::ProgressiveFrames:
      img->IsFramePicture = true;
      break;
    case InputPreprocessor_SingleFrames::Parameter::FieldsInFrames:
      img->Is2FieldPicture = true;
      img->TopFieldFirst   = d_parameter.m_TopFieldFirst;
      img->SkipFirstField  = false;
      img->SkipLastField   = false;
      break;
    case InputPreprocessor_SingleFrames::Parameter::Fields:
      img->IsFieldPicture = true;
      img->IsTopField = ((((d_next_num - d_parameter.m_first_num)%2)==0) == d_parameter.m_TopFieldFirst);
      break;
    }

  d_next_num++;
  GenerateFilename();

  return img;
}

void InputPreprocessor_SingleFrames::GenerateFilename()
{
  sprintf(d_next_file,d_filename_template,d_next_num);
}

