/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#ifndef INPUT_PREPROC_SGLFRMS_HH
#define INPUT_PREPROC_SGLFRMS_HH

#include "types.hh"
#include "input/preproc.hh"
#include "image/image.hh"

const int MaxFilenameLen = 100;  // Fuer dumme Compiler auch ausserhalb der Klasse!


class InputPreprocessor_SingleFrames : public InputPreprocessor
{
public:
  struct Parameter
  {
    Parameter();

    enum InputPicType { ProgressiveFrames, FieldsInFrames, Fields } m_InputType;
    enum InputFormat  { Format_PPM, Format_TGA } m_InputFormat;
    bool m_TopFieldFirst;
    int  m_first_num;
    int  m_last_num;
    int  m_framerate100;  // Framerate*100
  };

   InputPreprocessor_SingleFrames(const Parameter&,const char* nametemplate);
  ~InputPreprocessor_SingleFrames() { }

  bool         IsProgressiveSequence() const;
  bool         HasMoreInput() const;
  int          AskNextImageNum() const { return d_next_num; }
  const char*  AskNextImageFilename() const { return d_next_file; }
  class Image<Pixel>* GetNextImage(class Image<Pixel>*);
  int          AskFramerate100() const { return d_parameter.m_framerate100; }

  int AskWidth() const  { return d_width; }
  int AskHeight() const { return d_height; }

private:
  Parameter   d_parameter;
  int         d_next_num;
  const char* d_filename_template;
  char        d_next_file[MaxFilenameLen];

  Image<Pixel>* d_nextimg;
  int d_width,d_height;

  void GenerateFilename();
};

#endif
