/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#ifndef MATH_DCT_HH
#define MATH_DCT_HH

#include "types.hh"
#include "image/codedimg.hh"

#ifndef DCTFloat
typedef double DCTFloat;
#endif


typedef int FDCTInt;
typedef int IDCTInt;

/*
const int FDCT_M1Scale      = 31;
const int FDCT_PostM1_VK    = 12;
const int FDCT_PostM1_NK    = 15;
const int FDCT_PostM1_Shift = 16;

const int FDCT_M2Scale      = 31;
const int FDCT_PostM2_VK    = 16;
const int FDCT_PostM2_NK    = 15;
const int FDCT_PostM2_Shift = 31;

const int FDCT_M3Scale      = 31;
const int FDCT_PostM3_VK    = 17;
const int FDCT_PostM3_NK    = 15;
const int FDCT_PostM3_Shift = 31;
*/
const int FDCT_M1Scale      = 30;
const int FDCT_PostM1_VK    = 12;
const int FDCT_PostM1_NK    = 16;
const int FDCT_PostM1_Shift = 14;

const int FDCT_M2Scale      = 30;
const int FDCT_PostM2_VK    = 16;
const int FDCT_PostM2_NK    = 15;
const int FDCT_PostM2_Shift = 31;

const int FDCT_M3Scale      = 30;
const int FDCT_PostM3_VK    = 17;
const int FDCT_PostM3_NK    = 15;
const int FDCT_PostM3_Shift = 30;


/*
const int IDCT_M3Scale      = 33;
const int IDCT_PostM3_VK    = 10;
const int IDCT_PostM3_NK    = 19;
const int IDCT_PostM3_Shift = 14;

const int IDCT_M2Scale      = 29;
const int IDCT_PostM2_VK    = 13;
const int IDCT_PostM2_NK    = 13;
const int IDCT_PostM2_Shift = 35;

const int IDCT_M1Scale      = 29;
const int IDCT_PostM1_VK    = 13;
const int IDCT_PostM1_NK    = 13;
const int IDCT_PostM1_Shift = 29;
*/
const int IDCT_M3Scale      = 33;  // DOCS !
const int IDCT_PostM3_VK    = 10;
const int IDCT_PostM3_NK    = 20;
const int IDCT_PostM3_Shift = 13;

const int IDCT_M2Scale      = 28;
const int IDCT_PostM2_VK    = 12;
const int IDCT_PostM2_NK    = 16;
const int IDCT_PostM2_Shift = 32;

const int IDCT_M1Scale      = 28;
const int IDCT_PostM1_VK    = 16;
const int IDCT_PostM1_NK    = 14;
const int IDCT_PostM1_Shift = 30;


#define SHOW_M1 0
#define SHOW_M2 0
#define SHOW_M3 0

extern int ibuf[8][8];
extern int fbuf[8][8];

class FDCT1
{
public:
  virtual ~FDCT1() { }

  virtual void SetOutputScalefactors(const DCTFloat fact[8]) = 0;
  virtual void RemoveOutputScalefactors() = 0;

  virtual void PerformFDCT1(const DCTFloat in[8],DCTFloat out[8]) = 0;

protected:
  void SetScaleFact(DCTFloat fact[8]) { SetOutputScalefactors(fact); }
  void RemScaleFact()                 { RemoveOutputScalefactors();  }
};

class IDCT1
{
public:
  virtual ~IDCT1() { }

  virtual void SetOutputScalefactors(const DCTFloat fact[8]) = 0;
  virtual void RemoveOutputScalefactors() = 0;

  virtual void PerformIDCT1(const DCTFloat in[8],DCTFloat out[8]) = 0;

protected:
  void SetScaleFact(DCTFloat fact[8]) { SetOutputScalefactors(fact); }
  void RemScaleFact()                 { RemoveOutputScalefactors();  }
};


class FDCT2
{
public:
  virtual ~FDCT2() { }

  virtual void SetOutputScalefactors(const DCTFloat* fact[8] /* [y][x] */ ) = 0;
  virtual void RemoveOutputScalefactors() = 0;

  virtual void PerformFDCT2(const short* in[8] /* [y][x] */, DCTBlock& out) = 0;
          void PerformFDCT2(const uint8* in[8] /* [y][x] */, DCTBlock& out)
    {
      short  f[8][8];
      short* fp[8]; {for (int i=0;i<8;i++) fp[i]=f[i];}
      for (int i=0;i<8;i++)
        for (int j=0;j<8;j++)
          fp[i][j]=in[i][j];

      PerformFDCT2((const short**)fp,out);
    }

protected:
  void SetScaleFact(const DCTFloat* fact[8]) { SetOutputScalefactors(fact); }
  void RemScaleFact()                        { RemoveOutputScalefactors();  }
};

class IDCT2
{
public:
  virtual ~IDCT2() { }

  virtual void SetOutputScalefactors(const DCTFloat* fact[8] /* [v][u] */) = 0;
  virtual void RemoveOutputScalefactors() = 0;

  virtual void PerformIDCT2(const short* in[8] /* [v][u] */,uint8* out[8] /* [y][x] */,bool add) = 0;
          void PerformIDCT2Blk(const DCTBlock& in,uint8* out[8] /* [y][x] */, bool add)
    {
      const short* inp[8];
      for (int i=0;i<8;i++) 
        { inp[i] = in.m_coeff[i]; }
      PerformIDCT2(inp,out,add);
    }

protected:
  void SetScaleFact(const DCTFloat* fact[8]) { SetOutputScalefactors(fact); }
  void RemScaleFact()                        { RemoveOutputScalefactors();  }
};

#endif
