/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "math/dct1_theoretic.hh"
#include <math.h>

#ifndef PI
#define PI 3.1415926535
#endif

static DCTFloat costab[8][8]; // Tabelle aller benoetigten Cosinus-Werte
static DCTFloat C0; // 1/sqrt(2)

static class InitCosTab  // Dummy-Klasse zur Initialisierung der Cosinuswerte-Tabelle
{
public:
  InitCosTab()
  {
    for (int x=0;x<8;x++)
      for (int u=0;u<8;u++)
	costab[x][u] = cos((2*((DCTFloat)x)+1)*((DCTFloat)u)*PI/16.0);

    C0 = sqrt(0.5);
  }
} initcostab;


void DCT1_Theoretic::SetOutputScalefactors(DCTFloat fact[8])
{
  for (int i=0;i<8;i++)
    d_output_factors[i] = fact[i];

  d_scale_output=true;
}

void DCT1_Theoretic::PerformFDCT1(const DCTFloat in[8],DCTFloat out[8])
{
  for (int u=0;u<8;u++)
    {
      DCTFloat sum=0.0;

      for (int x=0;x<8;x++)
	sum += in[x] * costab[x][u];

      if (u==0) sum *= C0*0.5;
      else      sum *=    0.5;

      if (d_scale_output)
	sum *= d_output_factors[u];

      out[u] = sum;
    }
}


void DCT1_Theoretic::PerformIDCT1(const DCTFloat in[8],DCTFloat out[8])
{
  DCTFloat IN[8];

  if (d_scale_output)
    for (int i=0;i<8;i++)
      IN[i] = in[i]/d_output_factors[i];
  else
    for (int i=0;i<8;i++)
      IN[i] = in[i];

   
  for (int x=0;x<8;x++)
    {
      DCTFloat sum=0.0;

      for (int u=0;u<8;u++)
	{
	  DCTFloat val = IN[u] * costab[x][u];
	  if (u==0) val *= C0*0.5;
	  else      val *=    0.5;
	  sum += val;
	}

      out[x] = sum;
    }
}
