/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/
/*
  1-D DCT direkte Implementation der theoretischen Gleichungen

  Ziemlich langsam, da 72 Multiplikationen (darunter 7 Shifts)
  benoetigt werden. Die optionale Ausgabeskalierung benoetigt nochmals
  8 Multiplikationen.
  Die IDCT benoetigt sogar 128 Multiplikationen (darunter 56 Shifts).
 */

#ifndef MATH_DCT1_THEORETIC_HH
#define MATH_DCT1_THEORETIC_HH

#include "types.hh"
#include "math/dct.hh"


class DCT1_Theoretic : public FDCT1, public IDCT1
{
public:
   DCT1_Theoretic() : d_scale_output(false) { }
  ~DCT1_Theoretic() { }

  void SetOutputScalefactors(DCTFloat fact[8]);
  void RemoveOutputScalefactors() { d_scale_output=false; }

  void PerformFDCT1(const DCTFloat in[8],DCTFloat out[8]);
  void PerformIDCT1(const DCTFloat in[8],DCTFloat out[8]);

private:
  bool     d_scale_output;
  DCTFloat d_output_factors[8];
};

#endif
