/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "mpeg/syntax.hh"
#include "mpeg/constants.hh"
#include "mpeg/utility.hh"
#include "mpeg/vlctabs.hh"
#include "options.hh"
#include <stdlib.h>

#include <iostream.h>
#include <iomanip.h>

void MPEGSyntaxCoder::WriteSequenceHeader()
{
  // ===== SequenceHeader =====

  d_bs->FillByte();
  d_bs->WriteBits(STARTCODE_SEQUENCE_HEADER,32);
  d_bs->WriteBits(d_seqhdr.m_width  & MASK12,12);
  d_bs->WriteBits(d_seqhdr.m_height & MASK12,12);
  d_bs->WriteBits(d_seqhdr.m_aspect_ratio_information,4);

  uint4 frame_rate_code;
  uint2 frame_rate_extension_n;
  uint5 frame_rate_extension_d;

  CodeFrameRate(d_seqhdr.m_framerate100,options.CreateMPEG2,
		frame_rate_code, frame_rate_extension_n, frame_rate_extension_d);

  d_bs->WriteBits(frame_rate_code,4);
  d_bs->WriteBits(d_seqhdr.m_bit_rate & MASK18,18);
  d_bs->WriteBits(1,1);
  d_bs->WriteBits(d_seqhdr.m_vbv_buffer_size & MASK10,10);
  d_bs->WriteBits(d_seqhdr.m_constrained_parameters_flag ? 1 : 0 ,1);

  if (d_seqhdr.m_load_intra_quantiser_matrix)
    {
      d_bs->WriteBits(1,1);
      for (int i=0;i<64;i++)
	d_bs->WriteBits(d_seqhdr.m_intra_matrix[i],8);
    }
  else
    d_bs->WriteBits(0,1);

  if (d_seqhdr.m_load_nonintra_quantiser_matrix)
    {
      d_bs->WriteBits(1,1);
      for (int i=0;i<64;i++)
	d_bs->WriteBits(d_seqhdr.m_nonintra_matrix[i],8);
    }
  else
    d_bs->WriteBits(0,1);


  // ===== SequenceHeader-Extension =====

  if (options.CreateMPEG2)
    {
      d_bs->FillByte();
      d_bs->WriteBits(STARTCODE_EXTENSION_START,32);
      d_bs->WriteBits(EXTID_SequenceExtension  , 4);
      d_bs->WriteBits(CodeProfileLevel(d_seqhdr.m_profile,d_seqhdr.m_level),8);
      d_bs->WriteBits(d_seqhdr.m_progressive_sequence ? 1 : 0 , 1);
      d_bs->WriteBits(d_seqhdr.m_chroma_format,2);
      d_bs->WriteBits((d_seqhdr.m_width >>12) & MASK2 , 2);
      d_bs->WriteBits((d_seqhdr.m_height>>12) & MASK2 , 2);
      d_bs->WriteBits((d_seqhdr.m_bit_rate>>18) & MASK12,12);
      d_bs->WriteBits(1,1);
      d_bs->WriteBits((d_seqhdr.m_vbv_buffer_size>>10) & MASK8,8);
      d_bs->WriteBits(d_seqhdr.m_low_delay ? 1 : 0,1);
      d_bs->WriteBits(frame_rate_extension_n,2);
      d_bs->WriteBits(frame_rate_extension_d,5);
    }
}

GOPHeaderData::GOPHeaderData()
{
  m_time_code_drop_frame_flag = false;
  m_time_code_hours    = 0;
  m_time_code_minutes  = 0;
  m_time_code_seconds  = 0;
  m_time_code_pictures = 0;
  m_closed_gop  = false;
  m_broken_link = false;
}

void GOPHeaderData::AdvanceOnePicture(int fps)
{
  if (++m_time_code_pictures == fps)
    {
      m_time_code_pictures=0;
      if (++m_time_code_seconds == 60)
	{
	  m_time_code_seconds=0;
	  if (++m_time_code_minutes == 60)
	    {
	      m_time_code_minutes=0;
	      if (++m_time_code_hours == 24)
		m_time_code_hours=0;
	    }
	}
    }
}

void MPEGSyntaxCoder::WriteGOPHeader(const GOPHeaderData& gd)
{
  d_bs->FillByte();
  d_bs->WriteBits(STARTCODE_GROUP_START,32);
  d_bs->WriteBits(gd.m_time_code_drop_frame_flag ? 1 : 0 ,1);
  d_bs->WriteBits(gd.m_time_code_hours,5);
  d_bs->WriteBits(gd.m_time_code_minutes,6);
  d_bs->WriteBits(1,1);
  d_bs->WriteBits(gd.m_time_code_seconds,6);
  d_bs->WriteBits(gd.m_time_code_pictures,6);
  d_bs->WriteBits(gd.m_closed_gop  ? 1 : 0 ,1);
  d_bs->WriteBits(gd.m_broken_link ? 1 : 0 ,1);
}


void MPEGSyntaxCoder::WritePictureHeader(const PictureHeaderData& pd)
{
  d_bs->FillByte();
  d_bs->WriteBits(STARTCODE_PICTURE_START,32);
  d_bs->WriteBits(pd.m_temporal_reference,10);
  d_bs->WriteBits(pd.m_picture_coding_type,3);
  d_bs->WriteBits(pd.m_vbv_delay,16);

  if (pd.m_picture_coding_type == PICTYPE_P ||
      pd.m_picture_coding_type == PICTYPE_B)
    {
      if (!options.CreateMPEG2)
        { assert(pd.m_f_code[0][0] == pd.m_f_code[0][1]); }

      d_bs->WriteBits(pd.m_full_pel_forward_vector ? 1 : 0 ,1);
      d_bs->WriteBits(pd.m_f_code[0][0],3);
    }

  if (pd.m_picture_coding_type == PICTYPE_B)
    {
      if (!options.CreateMPEG2)
        { assert(pd.m_f_code[1][0] == pd.m_f_code[1][1]); }

      d_bs->WriteBits(pd.m_full_pel_backward_vector ? 1 : 0 ,1);
      d_bs->WriteBits(pd.m_f_code[1][0],3);
    }

  d_bs->WriteBits(0,1);

  if (options.CreateMPEG2)
    {
      d_bs->FillByte();
      d_bs->WriteBits(STARTCODE_EXTENSION_START,32);
      d_bs->WriteBits(EXTID_PictureCodingExtension , 4);
      d_bs->WriteBits(pd.m_f_code[0][0],4);
      d_bs->WriteBits(pd.m_f_code[0][1],4);
      d_bs->WriteBits(pd.m_f_code[1][0],4);
      d_bs->WriteBits(pd.m_f_code[1][1],4);
      d_bs->WriteBits(pd.m_intra_dc_precision-8,2);
      d_bs->WriteBits(pd.m_picture_structure,2);
      d_bs->WriteBits(pd.m_top_field_first ? 1:0 ,1);
      d_bs->WriteBits(pd.m_frame_pred_frame_dct ? 1:0 ,1);
      d_bs->WriteBits(pd.m_concealment_motion_vectors ? 1:0 ,1);
      d_bs->WriteBits(pd.m_q_scale_type,1);
      d_bs->WriteBits(pd.m_intra_vlc_format,1);
      d_bs->WriteBits(pd.m_alternate_scan?1:0 ,1);
      d_bs->WriteBits(pd.m_repeat_first_field?1:0 ,1);

      if (d_seqhdr.m_chroma_format == CHROMA_420)
	d_bs->WriteBits(pd.m_progressive_frame?1:0 ,1);
      else
	d_bs->WriteBits(0,1);

      d_bs->WriteBits(pd.m_progressive_frame?1:0 ,1);
      d_bs->WriteBits(pd.m_composite_display_flag?1:0 ,1);

      if (pd.m_composite_display_flag)
	{
	  d_bs->WriteBits(pd.m_v_axis,1);
	  d_bs->WriteBits(pd.m_field_sequence,3);
	  d_bs->WriteBits(pd.m_sub_carrier?1:0 ,1);
	  d_bs->WriteBits(pd.m_burst_amplitude,7);
	  d_bs->WriteBits(pd.m_sub_carrier_phase,8);
	}
    }
  else
    {
      assert(pd.m_intra_dc_precision==8);
      assert(pd.m_picture_structure==PICSTRUCT_FramePicture);
      assert(pd.m_frame_pred_frame_dct==true);
      assert(pd.m_q_scale_type==0);
      assert(pd.m_intra_vlc_format==0);
      assert(pd.m_alternate_scan==false);
    }
}

void MPEGSyntaxCoder::WriteSliceHeader(const SliceHeaderData& sd)
{
  // sleep(1);
  // cerr << "SLICE " << sd.m_slice_vertical_position << endl;


  d_bs->FillByte();
  d_bs->WriteBits(0x000001,24);

  if (d_seqhdr.m_height<=2800)
    {
      d_bs->WriteBits(sd.m_slice_vertical_position+1,8);
    }
  else
    {
      uint3 low7bits = sd.m_slice_vertical_position&MASK7;
      d_bs->WriteBits(low7bits,8);
      d_bs->WriteBits((sd.m_slice_vertical_position>>7)&MASK3,3);
    }

  // TODO: jetzt kommen eigentlich die Scalability-Felder

  d_bs->WriteBits(sd.m_quantiser_scale_code,5);
  if (sd.m_CodeIntraSlice)
    {
      d_bs->WriteBits(1,1);
      d_bs->WriteBits(sd.m_intra_slice ? 1:0 ,1);
      d_bs->WriteBits(0,7);
    }
  
  d_bs->WriteBits(0,1);
}

void MPEGSyntaxCoder::WritePicture(ImageData& img)
{
  // m_CodeBlk[] reduzieren (auf 'false' setzen, falls alle Koeffizienten 0 sind).

  if (img.pichdr.m_picture_coding_type != PICTYPE_I)  // keine uncodierten Bloecke I-Frames (CBP)
    {
      for (int y=0;y<img.mblks.AskHeight();y++)
        for (int x=0;x<img.mblks.AskWidth();x++)
          {
            Macroblock& mb = img.mblks.Ask(y,x);
            
            for (int i=0;i<12;i++)
              {
                if (mb.m_CodeBlk[i])
                  for (int v=0;v<8;v++)
                    for (int u=0;u<8;u++)
                      if (mb.m_blks[i].m_coeff[v][u] != 0) { goto break0; }
                
                mb.m_CodeBlk[i]=false;
break0:
                ;
              }
            
            mb.m_HasData = false;
            for (int j=0;j<12 && !mb.m_HasData;j++)
              mb.m_HasData |= mb.m_CodeBlk[j];
          }
    }



  // Skipped-Makrobloecke erzeugen

  if (img.pichdr.m_picture_coding_type != PICTYPE_I)  // keine Skippted-MBs in I-Frames
    {
      Macroblock* lastmb = NULL;
      Macroblock* mb;

      for (int y=0;y<img.mblks.AskHeight();y++)
        for (int x=0;x<img.mblks.AskWidth();x++ , lastmb=mb)
          {
            mb = &img.mblks.Ask(y,x);

            // Erster und letzter MB duerfen nicht uebersprungen werden.
            if (mb->m_BeginSlice || mb->m_EndSlice)
              continue;

            // MB hat Differenzdaten -> nicht ueberspringen.
            if (mb->m_HasData)
              continue;

            if (img.pichdr.m_picture_coding_type == PICTYPE_P)
              {
                if (!mb->m_IsIntra &&
                    mb->m_forward1.m_habs == 0 &&
                    mb->m_forward1.m_vabs == 0)
                  {
                    mb->m_Skip=true;
                  }
              }
            else
              {
                if (!lastmb->m_IsIntra && !lastmb->m_IsIntra &&
                    mb->m_HasMotionForward  == lastmb->m_HasMotionForward &&
                    mb->m_HasMotionBackward == lastmb->m_HasMotionBackward)
                  {
                    if (mb->m_HasMotionForward && mb->m_HasMotionBackward &&
                        mb->m_forward1.m_habs  == lastmb->m_forward1.m_habs  &&
                        mb->m_forward1.m_vabs  == lastmb->m_forward1.m_vabs  &&
                        mb->m_backward1.m_habs == lastmb->m_backward1.m_habs &&
                        mb->m_backward1.m_vabs == lastmb->m_backward1.m_vabs)
                      {
                        mb->m_Skip=true;
                      }
                    else if (mb->m_HasMotionForward && !mb->m_HasMotionBackward &&
                             mb->m_forward1.m_habs == lastmb->m_forward1.m_habs  &&
                             mb->m_forward1.m_vabs == lastmb->m_forward1.m_vabs)
                      {
                        mb->m_Skip=true;
                      }
                    else if (!mb->m_HasMotionForward && mb->m_HasMotionBackward &&
                             mb->m_backward1.m_habs == lastmb->m_backward1.m_habs &&
                             mb->m_backward1.m_vabs == lastmb->m_backward1.m_vabs)
                      {
                        mb->m_Skip=true;
                      }
                  }
              }
          }
    }


  // Gemeinsame Daten berechnen

  for (int a=0;a<2;a++)
    for (int b=0;b<2;b++)
      {
	assert(img.pichdr.m_f_code[a][b] != 0);
	d_f[a][b]     = (1<<(img.pichdr.m_f_code[a][b]-1));
	d_range[a][b] = 16* d_f[a][b];
      }


  // codiertes Bild speichern

  WritePictureHeader(img.pichdr);

  uint5 current_qscale=0;

  short init_predictor = (128<<(img.pichdr.m_intra_dc_precision-8));

  short pred_y,pred_cb,pred_cr;
  pred_y=pred_cb=pred_cr = init_predictor;

  ResetPMV();


  bool CodingMB = false;

  int mb_incr=0; // NEU

  for (int y=0;y<img.mblks.AskHeight();y++)
    {
      if (options.CreateMPEG2)
	{
	  assert(CodingMB==false); // Slice darf sich bei MPEG2 nicht ueber mehrere MB-Rows erstrecken.
	}

      // WAR int mb_incr=1;

      for (int x=0;x<img.mblks.AskWidth();x++)
	{
          //cout << "MB: " << x+y*img.mblks.AskWidth() << endl;

	  const Macroblock& mb = img.mblks.Ask_const(y,x);

          if (mb.m_BeginSlice) // NEU
            mb_incr=1;

          assert(mb_incr>0);

	  if (mb.m_Skip)
	    {
	      mb_incr++;
	      continue;
	    }
	  
	  /* CodeXX[] gibt an, ob der zugehoerige DCT-Block codiert werden soll. Dieses Flag
	     entspricht mb.m_CodeXX[], wird aber nur dann auf 'true' gesetzt, falls nicht alle
	     Koeffizienten 0 sind.
	     Bei Intra-Bloecken wird der Koeffizient (0;0) nicht beachtet, d.h. auch wenn nur
	     dieser ungleich 0 ist, wird CodeXX[] auf false gesetzt.
	     */

	  bool CodeY[4];
	  bool CodeCb[4];
	  bool CodeCr[4];

	  if (mb.m_IsIntra)
	    for (int i=0;i<4;i++)
	      {
		CodeY [i] = mb.m_CodeY [i];
		CodeCb[i] = mb.m_CodeCb[i];
		CodeCr[i] = mb.m_CodeCr[i];
	      }
	  else
	    for (int i=0;i<4;i++)
	      {
		CodeY[i] = CodeCb[i] = CodeCr[i] = false;
		
		if (mb.m_CodeY[i])
		  for (int v=0;v<8;v++)
		    for (int u=0;u<8;u++)
		      if (mb.m_y[i].m_coeff[v][u] != 0) { CodeY[i]=true; goto break1; }
break1:
		if (mb.m_CodeCb[i])
		  for (int v=0;v<8;v++)
		    for (int u=0;u<8;u++)
		      if (mb.m_cb[i].m_coeff[v][u] != 0) { CodeCb[i]=true; goto break2; }
break2:
		if (mb.m_CodeCr[i])
		  for (int v=0;v<8;v++)
		    for (int u=0;u<8;u++)
		      if (mb.m_cr[i].m_coeff[v][u] != 0) { CodeCr[i]=true; goto break3; }
break3:
		;
	    }

	  if (mb.m_BeginSlice)
	    {
	      assert(CodingMB==false);
	      CodingMB=true;
	      
	      // --- SliceHeader zusammenbasteln ---

	      SliceHeaderData sd;
	      sd.m_slice_vertical_position = y;
	      // sd.m_priority_breakpoint = 0;   // derzeit unbenutzt

              sd.m_quantiser_scale_code = current_qscale = mb.m_quantiser_scale_code;

	      sd.m_CodeIntraSlice = options.mpeg2.CodeIntraSliceFlag;

	      if (sd.m_CodeIntraSlice)
		{
		  // Schauen, ob Slice nur Intra-Bloecke enthaelt

		  bool  AllIntra = true;
		  for (int xx=x;xx<img.mblks.AskWidth() && AllIntra;xx++)
		    if (!img.mblks.Ask_const(y,xx).m_IsIntra)
		      AllIntra=false;
		  sd.m_intra_slice = AllIntra;
		}

	      WriteSliceHeader(sd);
	      
	      pred_y=pred_cb=pred_cr = init_predictor;
	      ResetPMV();
	    }
	  
	  // ------- Makroblock codieren ---------

          //cerr << "MAKROBLOCK " << y*img.mblks.AskWidth()+x << endl;

	  if (mb_incr>1)
	    {
	      pred_y=pred_cb=pred_cr = init_predictor;  /* Reset, wenn AddrIncr>1 */

              if (img.pichdr.m_picture_coding_type == PICTYPE_P)
                ResetPMV();
	    }

	  WriteMacroblockAddressIncrement(*d_bs,mb_incr);
	  mb_incr=1;

	  /* Syntaxelement: macroblock_modes() */

	  // MBmodes bestimmen und codieren

	  bool mb_quant,mb_mforward,mb_mbackward,mb_pattern,mb_intra; // ,mb_sptempwghtflag=false;
	  
	  mb_quant     = (mb.m_quantiser_scale_code != current_qscale);
	  mb_intra     = mb.m_IsIntra;
	  mb_mforward  = mb.m_HasMotionForward;
	  mb_mbackward = mb.m_HasMotionBackward;

	  if (mb_intra)
	    mb_pattern=false;
	  else
	    {
	      bool DCTCoded=false;
	      for (int i=0;i<4;i++) DCTCoded = DCTCoded | CodeY[i] | CodeCb[i] | CodeCr[i];
	      mb_pattern = DCTCoded;
	    }

	  if (!mb_pattern && !mb_intra) mb_quant=false;

#if 0
          if (!mb_mforward && !mb_mbackward)
            {
              cerr << "BUGFIX 1 used !\n";
              mb_pattern=false;
              mb_intra  =false;
            }
#endif

          if (img.pichdr.m_picture_coding_type == PICTYPE_B)
            {
              if (!mb_mforward && !mb_mbackward && mb_pattern && !mb_intra)
                {
                  cerr << "BUGFIX 2 used !\n";
                  mb_pattern=false;
                  mb_intra  =true;
                }
            }

	  uint8 packedmodes=0;
	  if (mb_quant)     packedmodes |= 0x10;
	  if (mb_mforward)  packedmodes |= 0x08;
	  if (mb_mbackward) packedmodes |= 0x04;
	  if (mb_pattern)   packedmodes |= 0x02;
	  if (mb_intra)     packedmodes |= 0x01;
	  // if (mb_sptempwghtflag) packedmodes |= 0x01;

          if (mb_intra)
            {
              assert(!mb_mforward);
              assert(!mb_mbackward);
              assert(!mb_pattern);
            }

          if (mb_quant)
            {
              assert(mb_pattern || mb_intra);
            }

	  WriteMacroblockModes(*d_bs , packedmodes,
			       img.pichdr.m_picture_coding_type,0 /* unused */);
	  if (mb_mforward || mb_mbackward)
	    {
	      if (img.pichdr.m_picture_structure == PICSTRUCT_FramePicture)
		{
		  if ( ! img.pichdr.m_frame_pred_frame_dct )
		    d_bs->WriteBits(mb.m_frame_motion_type,2);
		}
	      else
		{
		  d_bs->WriteBits(mb.m_field_motion_type,2);
		}
	    }

	  if (img.pichdr.m_picture_structure == PICSTRUCT_FramePicture &&
	      ! img.pichdr.m_frame_pred_frame_dct &&
	      (mb_intra || mb_pattern))
	    {
              d_bs->WriteBits(mb.m_FieldDCT ? 1 : 0 ,1);
            }

	  /* Ende des Syntaxelements: macroblock_modes() */

	  
	  if (mb_quant)
	    {
	      d_bs->WriteBits(mb.m_quantiser_scale_code,5);
	      current_qscale = mb.m_quantiser_scale_code;
	    }

	  if (mb_mforward || (mb_intra && img.pichdr.m_concealment_motion_vectors))
	    {
	      motion_vectors(img.pichdr, mb, 0);
	    }

	  if (mb_mbackward)
	    {
	      motion_vectors(img.pichdr, mb, 1);
	    }

	  if (mb_intra && img.pichdr.m_concealment_motion_vectors)
	    {
	      d_bs->WriteBits(1,1);
	    }


	  /* CBP */

	  if (mb_pattern)
	    {
	      uint8 cbr1=0,cbr2=0;
	      if (CodeY[0])  cbr1 |= 0x20;
	      if (CodeY[1])  cbr1 |= 0x10;
	      if (CodeY[2])  cbr1 |= 0x08;
	      if (CodeY[3])  cbr1 |= 0x04;
	      if (CodeCb[0]) cbr1 |= 0x02;
	      if (CodeCr[0]) cbr1 |= 0x01;

	      if (d_seqhdr.m_chroma_format != CHROMA_420)
		{
		  if (CodeCb[2]) cbr1 |= 0x20;
		  if (CodeCr[2]) cbr1 |= 0x10;
		  if (CodeCb[1]) cbr1 |= 0x08;
		  if (CodeCr[1]) cbr1 |= 0x04;
		  if (CodeCb[3]) cbr1 |= 0x02;
		  if (CodeCr[3]) cbr1 |= 0x01;
		}


              //cout << ((void*)cbr1) << endl;
	      WriteCBP(*d_bs,cbr1,cbr2,d_seqhdr.m_chroma_format);
	    }

	  /* die DCT-Bloecke */

	  const bool AltScan = img.pichdr.m_alternate_scan;
	  const bool UseB15  = img.pichdr.m_intra_vlc_format;

	  for (int i=0;i<4;i++)
	    {
	      if (CodeY[i]) CodeBlock(mb.m_y[i],mb_intra,true,pred_y,AltScan,UseB15);
	    }

	  if (CodeCb[0]) CodeBlock(mb.m_cb[0],mb_intra,false,pred_cb,AltScan,UseB15);
	  if (CodeCr[0]) CodeBlock(mb.m_cr[0],mb_intra,false,pred_cr,AltScan,UseB15);
	  if (CodeCb[2]) CodeBlock(mb.m_cb[2],mb_intra,false,pred_cb,AltScan,UseB15);
	  if (CodeCr[2]) CodeBlock(mb.m_cr[2],mb_intra,false,pred_cr,AltScan,UseB15);
	  if (CodeCb[1]) CodeBlock(mb.m_cb[1],mb_intra,false,pred_cb,AltScan,UseB15);
	  if (CodeCr[1]) CodeBlock(mb.m_cr[1],mb_intra,false,pred_cr,AltScan,UseB15);
	  if (CodeCb[3]) CodeBlock(mb.m_cb[3],mb_intra,false,pred_cb,AltScan,UseB15);
	  if (CodeCr[3]) CodeBlock(mb.m_cr[3],mb_intra,false,pred_cr,AltScan,UseB15);

	  if (!mb_intra)
	    pred_y=pred_cb=pred_cr = init_predictor;

	  if (mb_intra)
	    ResetPMV();

	  if (mb.m_EndSlice)
	    {
	      assert(CodingMB==true);
	      CodingMB=false;
	    }
	}
    }
}

static int getlog2(int n)
{
  static int logtab[] = { 0,1,2,4,8,16,32,64,128,256,512,1024,2048 };

  assert(n<=2048);
  int i=0;
  while (logtab[i] <= n)
    i++;

  return i-1;
}

struct Scan { uint8 u,v; };
static Scan zigzag[64] =
{ { 0, 0 },  { 1, 0 },  { 0, 1 },  { 0, 2 },
  { 1, 1 },  { 2, 0 },  { 3, 0 },  { 2, 1 },
  { 1, 2 },  { 0, 3 },  { 0, 4 },  { 1, 3 },
  { 2, 2 },  { 3, 1 },  { 4, 0 },  { 5, 0 },
  { 4, 1 },  { 3, 2 },  { 2, 3 },  { 1, 4 },
  { 0, 5 },  { 0, 6 },  { 1, 5 },  { 2, 4 },
  { 3, 3 },  { 4, 2 },  { 5, 1 },  { 6, 0 },
  { 7, 0 },  { 6, 1 },  { 5, 2 },  { 4, 3 },
  { 3, 4 },  { 2, 5 },  { 1, 6 },  { 0, 7 },
  { 1, 7 },  { 2, 6 },  { 3, 5 },  { 4, 4 },
  { 5, 3 },  { 6, 2 },  { 7, 1 },  { 7, 2 },
  { 6, 3 },  { 5, 4 },  { 4, 5 },  { 3, 6 },
  { 2, 7 },  { 3, 7 },  { 4, 6 },  { 5, 5 },
  { 6, 4 },  { 7, 3 },  { 7, 4 },  { 6, 5 },
  { 5, 6 },  { 4, 7 },  { 5, 7 },  { 6, 6 },
  { 7, 5 },  { 7, 6 },  { 6, 7 },  { 7, 7 } };

static Scan alternate[64] = 
{ { 0, 0 },  { 0, 1 },  { 0, 2 },  { 0, 3 },
  { 1, 0 },  { 1, 1 },  { 2, 0 },  { 2, 1 },
  { 1, 2 },  { 1, 3 },  { 0, 4 },  { 0, 5 },
  { 0, 6 },  { 0, 7 },  { 1, 7 },  { 1, 6 },
  { 1, 5 },  { 1, 4 },  { 2, 3 },  { 2, 2 },
  { 3, 0 },  { 3, 1 },  { 4, 0 },  { 4, 1 },
  { 3, 2 },  { 3, 3 },  { 2, 4 },  { 2, 5 },
  { 2, 6 },  { 2, 7 },  { 3, 4 },  { 3, 5 },
  { 3, 6 },  { 3, 7 },  { 4, 2 },  { 4, 3 },
  { 5, 0 },  { 5, 1 },  { 6, 0 },  { 6, 1 },
  { 5, 2 },  { 5, 3 },  { 4, 4 },  { 4, 5 },
  { 4, 6 },  { 4, 7 },  { 5, 4 },  { 5, 5 },
  { 5, 6 },  { 5, 7 },  { 6, 2 },  { 6, 3 },
  { 7, 0 },  { 7, 1 },  { 7, 2 },  { 7, 3 },
  { 6, 4 },  { 6, 5 },  { 6, 6 },  { 6, 7 },
  { 7, 4 },  { 7, 5 },  { 7, 6 },  { 7, 7 } };

void MPEGSyntaxCoder::CodeBlock(const DCTBlock& blk,bool mbintra,bool isY,short& pred,
				bool alternatescan,bool UseB15)
{
#if 0
  for (int u=0;u<8;u++,cout << endl)
    for (int v=0;v<8;v++)
      {
        cout << setw(5) << blk.m_coeff[u][v];
      }
  int nbits=0;
#endif

  int coeff;

  if (mbintra)
    {
      short diff = blk.m_coeff[0][0] - pred;

      int size = getlog2(abs(diff));
      int diffbits;
      if (diff<0) diffbits = ~(-diff);
      else        diffbits =    diff;

      if (isY) { WriteDCSizeLumi  (*d_bs,size); }
      else     { WriteDCSizeChroma(*d_bs,size); }

      if (size>0)
	d_bs->WriteBitsMasked(diffbits,size);

      pred = blk.m_coeff[0][0];

      coeff=1;
    }
  else
    {
      coeff=0;
    }

  const Scan* scan = (alternatescan ? alternate : zigzag);

  int runlen=0;
  while (coeff<64)
    {
      short c = blk.m_coeff[ scan[coeff].v ][ scan[coeff].u ];

      if (c==0)
	runlen++;
      else
	{
	  // cout << "Coeff: " << coeff << "  Run: " << runlen << "  QVal: " << c << endl;
	  CodeRunValue(*d_bs,runlen,c, coeff==0, UseB15, !options.CreateMPEG2);

#if 0
          nbits += BitsForRunValue(runlen,c, UseB15, false);
#endif

	  runlen=0;
	}
      coeff++;
    }

  // if (runlen>0)
    {
      if (!UseB15) d_bs->WriteBits(2,2);
      else         d_bs->WriteBits(6,4);
    }

#if 0
  cout << nbits << endl;
  cout << endl;
#endif
}

int BitsToCodeCoeff(const Macroblock& mb,const PictureHeaderData& ph)
{
  const bool AltScan = ph.m_alternate_scan;
  const bool UseB15  = ph.m_intra_vlc_format;

  int bits = 0;
  
  const Scan* scan = (AltScan ? alternate : zigzag);

  for (int i=0;i<12;i++)
    if (mb.m_CodeBlk[i])
      {
	int coeff=1;   // DC-Coeffizienten nicht beachten, da das den Aufwand wahrscheinlich nicht lohnt.
	int runlen=0;
	while (coeff<64)
	  {
	    short c = mb.m_blks[i].m_coeff[ scan[coeff].v ][ scan[coeff].u ];
	    if (c==0)
	      runlen++;
	    else
	      {
		bits += BitsForRunValue(runlen,c, UseB15, false);
		runlen=0;
	      }
	    coeff++;
	  }
	
	if (!UseB15) bits+=2;
	else         bits+=4;
      }

  return bits;
}

void MPEGSyntaxCoder::motion_vectors(const PictureHeaderData& phdr,const Macroblock& mb,
				     int direction)
{
  if (mb.m_motion_vector_count == 1)
    {
      if (mb.m_mv_format == Macroblock::MVFormat_Field && mb.m_dmv!=1)
	{
	  d_bs->WriteBits(mb.m_vector[0][direction].m_vertical_field_select,1);
	}
      motion_vector(phdr,mb, 0,direction);
    }
  else
    {
      d_bs->WriteBits(mb.m_vector[0][direction].m_vertical_field_select,1);
      motion_vector(phdr,mb, 0,direction);
      
      d_bs->WriteBits(mb.m_vector[1][direction].m_vertical_field_select,1);
      motion_vector(phdr,mb, 1,direction);
    }
}

void MPEGSyntaxCoder::motion_vector (const PictureHeaderData& phdr,const Macroblock& mb,
				     int n,int direction)
{
  int hdiff = mb.m_vector[n][direction].m_habs - d_PMV[n][direction][0];
  if      (hdiff> d_range[direction][0]) hdiff -= 2*d_range[direction][0];
  else if (hdiff<-d_range[direction][0]) hdiff += 2*d_range[direction][0];

  CodeMotionCode(*d_bs,hdiff,d_f[direction][0],phdr.m_f_code[direction][0]-1);

  if (mb.m_dmv)
    { assert(0); /* TODO */ }

  int vdiff = mb.m_vector[n][direction].m_vabs - d_PMV[n][direction][1];
  if      (vdiff> d_range[direction][1]) vdiff -= 2*d_range[direction][1];
  else if (vdiff<-d_range[direction][1]) vdiff += 2*d_range[direction][1];

  CodeMotionCode(*d_bs,vdiff,d_f[direction][1],phdr.m_f_code[direction][1]-1);

  if (mb.m_dmv)
    { assert(0); /* TODO */ }

  // Den Praediktor neu setzen.

  d_PMV[n][direction][0] = mb.m_vector[n][direction].m_habs;
  d_PMV[n][direction][1] = mb.m_vector[n][direction].m_vabs;
}

void MPEGSyntaxCoder::ResetPMV()
{
  for (int r=0;r<2;r++)
  for (int s=0;s<2;s++)
  for (int t=0;t<2;t++)
    d_PMV[r][s][t] = 0;
}
