/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#ifndef OUTPUT_BITBUFFER_HH
#define OUTPUT_BITBUFFER_HH

#include "types.hh"
#include "output/streamsink.hh"

const int BufferLength = 16; // 1024*10;  // Muss Vielfaches von sizeof(BitArray) sein.


class BitBuffer
{
public:
   BitBuffer();
  ~BitBuffer();

  void WriteBits(BitArray bits,int nBits); // Bits sind rechtsbuendig drin.
  void WriteBitsMasked(BitArray bits,int nBits);
  void WriteBits(BitArray bits,int nBits,int firstbit); // Bits sind um 'firstbit' nach links verschoben.
  void FillByte(); // Auf die naechste Bytegrenze mit 0 auffuellen. Nichts machen, wenn schon an Grenze.
  void Flush();

  uint8* d_buffer;
  int    d_bufferidx;
private:
  int    d_buffersize;

  BitArray  d_tmpbuf;
  int       d_freebits;

  void TmpArrayToBuffer();

  static bool IsLittleEndian;
  friend class EndianTest;
};





const int BytesPerEntry = sizeof(BitArray);

inline void BitBuffer::TmpArrayToBuffer()
{
  if (d_bufferidx == d_buffersize)
    {
      uint8* newbuf = new uint8[d_buffersize*2];
      for (int i=0;i<d_buffersize;i++)
        newbuf[i]=d_buffer[i];
      
      d_buffersize *= 2;

      delete[] d_buffer;
      d_buffer = newbuf;
    }
  
  if (IsLittleEndian)
    {
      unsigned char* p = (unsigned char*)(&d_tmpbuf);
      for (int j=BytesPerEntry-1;j>=0;j--)
        d_buffer[d_bufferidx++] = p[j];
    }
  else
    {
      unsigned char* p = (unsigned char*)(&d_tmpbuf);
      for (int j=0;j<BytesPerEntry;j++)
        d_buffer[d_bufferidx++] = p[j];
    }
}

#endif
