/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "statistics/bitsestim.hh"
#include "mpeg/constants.hh"
#include "coding/utility.hh"
#include "options.hh"
#include <stdlib.h>

void CalcSpatialDerivative(const ImageData* img, Array2<uint32>& deriv)
{
  uint32 MB_Width  = img->mblks.AskWidth();
  uint32 MB_Height = img->mblks.AskHeight();

  deriv.Create(img->mblks.AskWidth(),
               img->mblks.AskHeight());

  const Pixel*const*     intra    = img->source->AskBitmap_const(Bitmap_Y).AskFrame_const();
  const DiffPixel*const* forward  = img->pred_forw.AskBitmap_const(Bitmap_Y).AskFrame_const();
  const DiffPixel*const* backward = NULL;
  const DiffPixel*const* bidir    = NULL;

  if (img->pichdr.m_picture_coding_type == PICTYPE_B)
    {
      backward = img->pred_backw.AskBitmap_const(Bitmap_Y).AskFrame_const();
      bidir    = img->pred_bidir.AskBitmap_const(Bitmap_Y).AskFrame_const();
    }

  for (uint32 y=0;y<MB_Height;y++)
    for (uint32 x=0;x<MB_Width;x++)
      {
        uint32 this_dh=0;
        uint32 this_dv=0;

        const Macroblock& mb = img->mblks.Ask_const(y,x);
        const DiffPixel*const* dpix = NULL;

             if (mb.m_HasMotionForward && mb.m_HasMotionBackward) dpix = bidir;
        else if (mb.m_HasMotionForward)                           dpix = forward;
        else if (mb.m_HasMotionBackward)                          dpix = backward;
        else
          {
            const Pixel*const* pix = intra;

            for (int yy=0;yy<16;yy++)
              for (int xx=0;xx<15;xx++)
                this_dh += abs(pix[y*16+yy][x*16+xx] - pix[y*16+yy][x*16+xx+1]);
            
            for (int yy=0;yy<15;yy++)
              for (int xx=0;xx<16;xx++)
                this_dv += abs(pix[y*16+yy][x*16+xx] - pix[y*16+yy+1][x*16+xx]);
          }

        if (dpix)
          {
            for (int yy=0;yy<16;yy++)
              for (int xx=0;xx<15;xx++)
                this_dh += abs(dpix[y*16+yy][x*16+xx] - dpix[y*16+yy][x*16+xx+1]);
            
            for (int yy=0;yy<15;yy++)
              for (int xx=0;xx<16;xx++)
                this_dv += abs(dpix[y*16+yy][x*16+xx] - dpix[y*16+yy+1][x*16+xx]);
          }

        deriv.Ask(y,x) = this_dh + this_dv;
      }
}

#define nDerivValues 21
int bittab_I[][nDerivValues] =
{
  /*  0*/  { /* DUMMY */ },
  /*  1*/  {   339,  902, 1055, 1274, 1439, 1546, 1647, 1743, 1829, 1927,
              2039, 2134, 2233, 2315, 2369, 2400, 2483, 2516, 2631, 2666, 2708 },
  /*  2*/  {   186,  523,  645,  814,  947, 1035, 1116, 1195, 1265, 1344,
              1432, 1509, 1591, 1660, 1696, 1720, 1784, 1814, 1909, 1918, 1958 },
  /*  3*/  {   111,  318,  416,  574,  682,  761,  835,  910,  974, 1042,
              1121, 1193, 1263, 1323, 1357, 1375, 1432, 1456, 1517, 1533, 1584 },
  /*  4*/  {    80,  221,  301,  442,  534,  605,  672,  741,  801,  860,
               934,  995, 1061, 1113, 1150, 1165, 1218, 1238, 1301, 1313, 1340 },
  /*  5*/  {    54,  141,  208,  333,  421,  486,  548,  611,  664,  720,
               786,  844,  905,  952,  985, 1000, 1048, 1070, 1124, 1141, 1170 },
  /*  6*/  {    44,  101,  155,  268,  354,  414,  473,  530,  578,  629,
               691,  743,  797,  844,  872,  889,  929,  957, 1009, 1016, 1038 },
  /*  7*/  {    36,   77,  122,  226,  309,  364,  418,  471,  516,  565,
               619,  669,  721,  764,  788,  804,  846,  865,  909,  926,  948 },
  /*  8*/  {    32,   60,   93,  187,  265,  316,  368,  418,  460,  506,
               558,  604,  653,  695,  717,  732,  769,  789,  826,  847,  884 },
  /* 10*/  {    26,   42,   62,  138,  207,  253,  298,  342,  379,  420,
               467,  508,  548,  588,  606,  626,  656,  673,  715,  719,  745 },
  /* 12*/  {    22,   33,   46,  107,  170,  212,  252,  291,  326,  362,
               403,  441,  479,  511,  529,  544,  574,  586,  624,  626,  650 },
  /* 14*/  {    20,   29,   39,   91,  147,  185,  221,  256,  287,  321,
               358,  393,  426,  455,  474,  485,  513,  527,  559,  562,  582 },
  /* 16*/  {    18,   26,   34,   80,  131,  165,  198,  230,  259,  290,
               324,  355,  386,  412,  431,  436,  463,  480,  506,  514,  532 },
  /* 18*/  {    17,   24,   30,   70,  114,  146,  176,  205,  231,  259,
               290,  318,  347,  372,  386,  395,  414,  436,  455,  466,  476 },
  /* 20*/  {    17,   22,   28,   63,  104,  133,  161,  189,  212,  239,
               267,  294,  320,  342,  359,  365,  382,  401,  417,  433,  445 },
  /* 22*/  {    16,   21,   26,   57,   93,  120,  146,  173,  194,  219,
               245,  272,  295,  316,  331,  337,  354,  372,  387,  400,  412 },
  /* 24*/  {    15,   20,   24,   53,   86,  111,  135,  160,  181,  204,
               228,  252,  276,  294,  308,  314,  327,  342,  366,  372,  382 },
  /* 28*/  {    15,   18,   22,   46,   75,   96,  118,  140,  159,  179,
               199,  221,  242,  260,  272,  275,  287,  305,  324,  328,  333 },
  /* 32*/  {    14,   17,   20,   41,   66,   86,  105,  125,  143,  160,
               179,  198,  217,  233,  245,  250,  259,  272,  293,  294,  301 },
  /* 36*/  {    14,   16,   19,   37,   59,   76,   93,  111,  127,  142,
               160,  177,  193,  209,  219,  222,  233,  243,  259,  261,  268 },
  /* 40*/  {    13,   15,   18,   34,   53,   68,   84,  100,  116,  129,
               145,  161,  175,  191,  200,  201,  213,  221,  234,  236,  245 },
  /* 44*/  {    13,   15,   17,   31,   48,   62,   76,   91,  106,  118,
               132,  147,  160,  175,  183,  185,  196,  204,  214,  217,  226 },
  /* 48*/  {    13,   15,   16,   29,   45,   58,   71,   85,   98,  109,
               123,  137,  148,  163,  169,  170,  183,  189,  199,  202,  213 },
  /* 52*/  {    13,   14,   15,   27,   42,   53,   65,   79,   91,  101,
               113,  126,  137,  149,  156,  159,  169,  176,  182,  183,  197 },
  /* 56*/  {    13,   14,   15,   25,   39,   49,   61,   73,   84,   94,
               105,  117,  127,  138,  145,  148,  158,  163,  171,  175,  180 },
  /* 64*/  {    12,   13,   14,   23,   35,   44,   54,   65,   75,   83,
                93,  103,  112,  122,  127,  131,  139,  143,  152,  152,  162 },
  /* 72*/  {    12,   13,   14,   21,   31,   39,   48,   58,   66,   73,
                82,   91,   99,  107,  112,  115,  121,  128,  136,  137,  143 },
  /* 80*/  {    12,   13,   13,   20,   28,   36,   43,   52,   60,   66,
                74,   82,   88,   96,  101,  104,  109,  115,  123,  124,  125 },
  /* 88*/  {    12,   13,   13,   18,   26,   33,   39,   48,   55,   60,
                67,   74,   80,   88,   91,   94,   99,  105,  109,  111,  112 },
  /* 96*/  {    12,   13,   13,   18,   25,   30,   37,   44,   50,   55,
                62,   68,   72,   81,   84,   87,   91,   97,   99,  100,  103 },
  /*104*/  {    12,   13,   13,   17,   23,   28,   34,   41,   46,   51,
                57,   62,   66,   74,   76,   79,   82,   88,   91,   92,   93 },
  /*112*/  {    12,   12,   13,   16,   22,   27,   32,   38,   44,   48,
                53,   57,   62,   69,   71,   72,   76,   81,   83,   85,   86 }
};


int bittab_F[][nDerivValues] =
{
  /*  0*/  { /* DUMMY */ },
  /*  1*/  {   932, 1381, 1713, 1938, 2158, 2329, 2441, 2547, 2633, 2728,
              2797, 2893, 2943, 3049, 3091, 3120, 3252, 3248, 3245, 3170, 3513 },
  /*  2*/  {   407,  597,  912, 1089, 1264, 1418, 1522, 1618, 1691, 1778,
              1840, 1922, 1970, 2071, 2086, 2159, 2238, 2258, 2218, 2175, 2445 },
  /*  3*/  {   221,  352,  573,  725,  870, 1002, 1097, 1184, 1249, 1332,
              1389, 1457, 1503, 1593, 1616, 1657, 1737, 1766, 1781, 1768, 1882 },
  /*  4*/  {   135,  222,  382,  518,  646,  763,  850,  930,  991, 1060,
              1117, 1181, 1226, 1303, 1330, 1362, 1427, 1457, 1458, 1419, 1611 },
  /*  5*/  {    84,  146,  252,  370,  488,  594,  672,  746,  804,  869,
               922,  978, 1019, 1090, 1124, 1152, 1197, 1245, 1237, 1192, 1395 },
  /*  6*/  {    58,  100,  172,  271,  377,  474,  550,  616,  673,  735,
               779,  839,  878,  941,  963,  998, 1056, 1072, 1061, 1054, 1150 },
  /*  7*/  {    43,   71,  118,  203,  298,  386,  455,  519,  572,  631,
               674,  729,  764,  825,  853,  873,  929,  940,  947,  958, 1044 },
  /*  8*/  {    34,   54,   82,  155,  240,  321,  386,  446,  496,  551,
               593,  643,  679,  728,  760,  801,  824,  843,  856,  852,  938 },
  /*  9*/  {    23,   36,   47,   96,  163,  231,  287,  341,  389,  436,
               474,  517,  554,  593,  617,  657,  673,  700,  706,  712,  772 },
  /* 10*/  {    18,   26,   32,   66,  119,  175,  225,  272,  314,  358,
               393,  431,  467,  507,  521,  551,  573,  598,  597,  640,  681 },
  /* 11*/  {    16,   21,   25,   49,   90,  137,  180,  221,  262,  299,
               333,  364,  399,  430,  454,  472,  503,  517,  537,  558,  573 },
  /* 12*/  {    15,   19,   21,   39,   72,  112,  148,  186,  221,  255,
               287,  318,  351,  381,  399,  415,  437,  461,  466,  480,  517 },
  /* 13*/  {    14,   17,   19,   33,   60,   93,  125,  158,  189,  222,
               250,  280,  309,  332,  350,  366,  401,  409,  411,  413,  459 },
  /* 14*/  {    14,   16,   17,   28,   51,   79,  107,  135,  166,  194,
               219,  248,  274,  297,  310,  329,  353,  371,  376,  360,  413 },
  /* 15*/  {    13,   15,   16,   25,   44,   68,   93,  117,  144,  171,
               195,  221,  246,  268,  279,  299,  324,  329,  336,  353,  382 },
  /* 16*/  {    13,   14,   15,   23,   39,   59,   81,  103,  127,  152,
               174,  197,  222,  245,  255,  280,  290,  307,  319,  291,  340 },
  /* 17*/  {    12,   14,   14,   20,   31,   47,   64,   83,  101,  122,
               143,  161,  182,  201,  218,  236,  250,  252,  261,  263,  286 },
  /* 18*/  {    12,   13,   13,   18,   26,   39,   53,   67,   83,   99,
               116,  134,  154,  170,  185,  194,  214,  223,  234,  233,  231 },
  /* 19*/  {    12,   13,   13,   16,   23,   33,   44,   56,   70,   82,
                98,  116,  129,  147,  160,  168,  183,  191,  199,  209,  225 },
  /* 20*/  {    12,   13,   13,   15,   21,   28,   38,   48,   59,   69,
                83,   99,  112,  123,  139,  149,  160,  162,  178,  186,  194 },
  /* 21*/  {    12,   12,   13,   14,   19,   25,   33,   42,   51,   60,
                71,   85,   96,  109,  119,  131,  142,  141,  155,  171,  187 },
  /* 22*/  {    12,   12,   12,   14,   18,   23,   30,   37,   45,   53,
                62,   73,   85,   98,  102,  111,  126,  127,  131,  153,  160 },
  /* 23*/  {    12,   12,   12,   13,   17,   21,   27,   34,   40,   47,
                56,   65,   74,   87,   89,  104,  105,  118,  115,  128,  147 },
  /* 24*/  {    12,   12,   12,   13,   16,   19,   24,   31,   36,   42,
                49,   56,   65,   79,   81,   90,  101,   99,  118,  113,  114 },
  /* 25*/  {    12,   12,   12,   13,   15,   17,   21,   26,   30,   35,
                40,   45,   52,   61,   67,   72,   80,   83,   92,  101,   97 },
  /* 26*/  {    12,   12,   12,   13,   14,   16,   19,   22,   26,   30,
                34,   39,   44,   51,   55,   62,   63,   70,   83,   70,   87 },
  /* 27*/  {    12,   12,   12,   12,   13,   15,   17,   20,   23,   26,
                29,   33,   38,   41,   45,   52,   57,   57,   64,   73,   77 },
  /* 28*/  {    12,   12,   12,   12,   13,   14,   16,   18,   21,   23,
                26,   29,   33,   38,   40,   40,   48,   48,   58,   57,   71 },
  /* 29*/  {    12,   12,   12,   12,   13,   14,   15,   17,   19,   21,
                24,   26,   29,   32,   35,   37,   40,   44,   48,   53,   58 },
  /* 30*/  {    12,   12,   12,   12,   13,   13,   15,   16,   18,   19,
                22,   24,   27,   28,   30,   32,   37,   37,   44,   41,   56 },
  /* 31*/  {    12,   12,   12,   12,   12,   13,   14,   15,   17,   19,
                20,   22,   24,   26,   27,   31,   31,   34,   40,   38,   37 }
};

int bittab_B[][nDerivValues] =
{
  /*  0*/  { /* DUMMY */ },
  /*  1*/  {  1040, 1421, 1710, 1932, 2157, 2328, 2442, 2547, 2629, 2719,
              2799, 2853, 2918, 2979, 3049, 3040, 3142, 3091, 3255, 3221, 3325 },
  /*  2*/  {   416,  679,  909, 1082, 1263, 1417, 1519, 1617, 1685, 1768,
              1836, 1889, 1958, 2000, 2060, 2085, 2142, 2116, 2193, 2245, 2314 },
  /*  3*/  {   225,  397,  573,  718,  868, 1000, 1094, 1177, 1244, 1325,
              1382, 1428, 1488, 1509, 1583, 1636, 1670, 1678, 1660, 1731, 1825 },
  /*  4*/  {   138,  255,  382,  511,  644,  760,  845,  921,  983, 1056,
              1111, 1159, 1198, 1244, 1298, 1344, 1389, 1361, 1442, 1422, 1510 },
  /*  5*/  {    85,  164,  251,  364,  486,  591,  670,  743,  797,  866,
               914,  951, 1003, 1036, 1089, 1128, 1161, 1155, 1264, 1200, 1357 },
  /*  6*/  {    57,  111,  171,  267,  376,  472,  547,  614,  664,  727,
               775,  811,  848,  876,  963,  963, 1023, 1031, 1043, 1017, 1194 },
  /*  7*/  {    41,   76,  118,  201,  298,  385,  455,  519,  569,  620,
               677,  704,  741,  780,  838,  870,  881,  867,  947,  923, 1026 },
  /*  8*/  {    31,   56,   82,  154,  240,  319,  385,  447,  494,  544,
               590,  629,  653,  703,  743,  763,  819,  762,  840,  827,  910 },
  /*  9*/  {    23,   35,   46,   97,  164,  229,  287,  339,  387,  433,
               468,  504,  536,  569,  605,  638,  679,  620,  670,  672,  782 },
  /* 10*/  {    18,   26,   31,   66,  119,  173,  224,  271,  317,  354,
               392,  422,  443,  481,  514,  547,  562,  539,  569,  596,  703 },
  /* 11*/  {    16,   21,   24,   50,   91,  135,  178,  220,  259,  297,
               332,  360,  383,  409,  446,  459,  513,  476,  509,  517,  576 },
  /* 12*/  {    15,   19,   21,   40,   72,  109,  146,  182,  219,  250,
               284,  316,  329,  358,  385,  408,  444,  424,  436,  479,  485 },
  /* 13*/  {    14,   17,   18,   33,   59,   90,  121,  153,  185,  218,
               246,  276,  291,  316,  348,  366,  396,  393,  389,  405,  397 },
  /* 14*/  {    13,   16,   17,   28,   49,   76,  102,  130,  159,  188,
               217,  241,  261,  282,  310,  318,  351,  356,  378,  332,  389 },
  /* 15*/  {    13,   15,   16,   25,   42,   64,   87,  111,  138,  166,
               188,  213,  238,  259,  272,  294,  329,  301,  319,  322,  302 },
  /* 16*/  {    13,   14,   15,   22,   37,   56,   76,   95,  122,  145,
               168,  188,  212,  237,  250,  265,  291,  310,  291,  294,  321 },
  /* 17*/  {    12,   14,   14,   19,   29,   43,   58,   75,   94,  116,
               134,  153,  166,  197,  212,  217,  263,  248,  257,  247,  280 },
  /* 18*/  {    12,   13,   13,   17,   24,   35,   47,   60,   75,   93,
               108,  127,  142,  159,  181,  197,  213,  209,  225,  236,  235 },
  /* 19*/  {    12,   13,   13,   15,   21,   29,   39,   50,   62,   75,
                91,  100,  123,  140,  148,  160,  188,  198,  197,  195,  228 },
  /* 20*/  {    12,   13,   13,   14,   19,   25,   34,   42,   51,   63,
                78,   85,  105,  115,  128,  138,  161,  179,  164,  212,  164 },
  /* 21*/  {    12,   12,   12,   14,   17,   22,   29,   36,   44,   55,
                63,   75,   85,   95,  114,  119,  135,  149,  155,  163,  147 },
  /* 22*/  {    12,   12,   12,   13,   16,   20,   26,   32,   38,   47,
                55,   66,   77,   82,   92,  103,  119,  120,  139,  155,  144 },
  /* 23*/  {    12,   12,   12,   13,   15,   18,   23,   29,   34,   41,
                48,   56,   69,   74,   80,   85,  112,  104,  136,  138,  140 },
  /* 24*/  {    12,   12,   12,   13,   14,   17,   22,   26,   30,   36,
                43,   48,   62,   61,   66,   80,   95,  101,  110,  120,  119 },
  /* 25*/  {    12,   12,   12,   13,   14,   15,   18,   22,   25,   29,
                35,   40,   46,   50,   62,   63,   72,   90,   91,   91,   90 },
  /* 26*/  {    12,   12,   12,   12,   13,   14,   17,   19,   22,   25,
                29,   33,   40,   43,   44,   50,   62,   72,   72,   76,   79 },
  /* 27*/  {    12,   12,   12,   12,   13,   14,   15,   17,   19,   22,
                25,   28,   34,   34,   37,   44,   47,   59,   51,   75,   74 },
  /* 28*/  {    12,   12,   12,   12,   13,   13,   15,   16,   18,   20,
                22,   25,   29,   29,   32,   36,   41,   42,   45,   57,   71 },
  /* 29*/  {    12,   12,   12,   12,   12,   13,   14,   15,   16,   18,
                20,   22,   25,   26,   27,   29,   39,   35,   41,   42,   60 },
  /* 30*/  {    12,   12,   12,   12,   12,   13,   14,   15,   16,   17,
                18,   20,   22,   25,   26,   27,   26,   36,   30,   34,   52 },
  /* 31*/  {    12,   12,   12,   12,   12,   13,   13,   14,   15,   16,
                17,   19,   20,   21,   26,   24,   26,   32,   34,   30,   42 }
};


int bittab_D[][nDerivValues] =
{
  /*  0*/  { /* DUMMY */ },
  /*  1*/  {   906, 1386, 1632, 1873, 2129, 2287, 2408, 2510, 2616, 2697,
              2779, 2830, 2927, 2977, 3069, 3062, 3065, 3120, 3217, 3220, 3123 },
  /*  2*/  {   335,  656,  851, 1034, 1241, 1379, 1486, 1580, 1677, 1751,
              1825, 1870, 1946, 1978, 2094, 2098, 2073, 2121, 2221, 2215, 2128 },
  /*  3*/  {   168,  371,  531,  683,  854,  977, 1072, 1156, 1238, 1309,
              1370, 1420, 1487, 1529, 1612, 1617, 1604, 1679, 1764, 1739, 1653 },
  /*  4*/  {    97,  222,  348,  482,  630,  739,  828,  902,  979, 1042,
              1105, 1142, 1204, 1246, 1311, 1319, 1330, 1351, 1410, 1445, 1352 },
  /*  5*/  {    57,  129,  224,  344,  476,  574,  652,  722,  795,  850,
               916,  952, 1009, 1036, 1106, 1118, 1153, 1116, 1225, 1209, 1124 },
  /*  6*/  {    37,   77,  146,  248,  369,  457,  532,  597,  665,  717,
               768,  811,  864,  893,  954,  957,  979,  983, 1064, 1043, 1005 },
  /*  7*/  {    27,   47,   95,  181,  288,  372,  442,  503,  568,  618,
               666,  697,  754,  794,  840,  860,  852,  856,  950,  954,  898 },
  /*  8*/  {    22,   31,   62,  134,  229,  308,  374,  433,  493,  543,
               583,  620,  672,  695,  743,  752,  784,  767,  848,  858,  792 },
  /*  9*/  {    17,   20,   32,   75,  149,  219,  280,  332,  385,  432,
               468,  503,  542,  568,  613,  635,  618,  650,  691,  738,  624 },
  /* 10*/  {    15,   16,   22,   48,  102,  162,  217,  266,  312,  354,
               388,  420,  449,  484,  512,  532,  549,  545,  588,  635,  548 },
  /* 11*/  {    14,   15,   18,   34,   73,  123,  174,  216,  260,  296,
               331,  360,  385,  417,  451,  456,  472,  451,  505,  545,  488 },
  /* 12*/  {    13,   14,   16,   27,   56,   96,  141,  179,  219,  255,
               285,  309,  334,  363,  391,  404,  419,  421,  440,  462,  465 },
  /* 13*/  {    13,   13,   15,   23,   45,   77,  116,  150,  186,  220,
               249,  273,  300,  320,  346,  361,  387,  368,  397,  428,  427 },
  /* 14*/  {    13,   13,   14,   20,   38,   63,   97,  128,  162,  193,
               220,  242,  265,  294,  312,  323,  341,  324,  358,  401,  356 },
  /* 15*/  {    13,   13,   14,   18,   32,   53,   82,  109,  140,  168,
               194,  216,  242,  255,  286,  297,  290,  308,  310,  371,  371 },
  /* 16*/  {    12,   13,   13,   17,   28,   46,   70,   95,  121,  149,
               171,  194,  216,  233,  256,  263,  275,  266,  292,  340,  327 },
  /* 17*/  {    12,   12,   13,   15,   23,   36,   53,   73,   94,  117,
               138,  155,  180,  194,  218,  221,  243,  232,  236,  285,  263 },
  /* 18*/  {    12,   12,   12,   14,   20,   29,   43,   58,   74,   92,
               114,  128,  149,  164,  183,  192,  206,  201,  199,  254,  254 },
  /* 19*/  {    12,   12,   12,   14,   18,   25,   35,   48,   61,   75,
                93,  107,  125,  143,  157,  166,  184,  179,  190,  233,  172 },
  /* 20*/  {    12,   12,   12,   13,   17,   22,   30,   40,   50,   63,
                78,   90,  105,  122,  138,  145,  153,  156,  180,  199,  185 },
  /* 21*/  {    12,   12,   12,   13,   16,   20,   27,   35,   43,   54,
                64,   74,   90,  106,  118,  124,  141,  139,  154,  180,  149 },
  /* 22*/  {    12,   12,   12,   13,   15,   19,   24,   30,   38,   47,
                54,   64,   73,   88,  102,  108,  128,  126,  131,  140,  174 },
  /* 23*/  {    12,   12,   12,   13,   14,   17,   22,   27,   33,   41,
                48,   57,   63,   74,   89,  100,  116,  111,  129,  137,  140 },
  /* 24*/  {    12,   12,   12,   12,   14,   16,   20,   25,   30,   36,
                45,   51,   56,   69,   80,   85,   98,  109,  109,  119,  126 },
  /* 25*/  {    12,   12,   12,   12,   13,   15,   18,   21,   25,   30,
                36,   41,   44,   54,   56,   68,   83,   81,   90,  102,  118 },
  /* 26*/  {    12,   12,   12,   12,   13,   14,   16,   19,   22,   25,
                30,   35,   37,   43,   51,   50,   66,   68,   76,   83,  101 },
  /* 27*/  {    12,   12,   12,   12,   13,   14,   15,   17,   19,   22,
                27,   30,   33,   35,   42,   43,   48,   60,   70,   60,   91 },
  /* 28*/  {    12,   12,   12,   12,   12,   13,   14,   16,   18,   20,
                24,   26,   29,   32,   35,   38,   41,   49,   60,   59,   62 },
  /* 29*/  {    12,   12,   12,   12,   12,   13,   14,   15,   17,   18,
                21,   23,   27,   27,   28,   33,   39,   38,   45,   53,   67 },
  /* 30*/  {    12,   12,   12,   12,   12,   13,   13,   14,   16,   17,
                19,   21,   23,   24,   24,   28,   33,   35,   38,   44,   43 },
  /* 31*/  {    12,   12,   12,   12,   12,   13,   13,   14,   15,   16,
                17,   20,   21,   23,   23,   27,   28,   30,   36,   41,   33 }
};


int Interpolate(int x1,int x2,int y1,int y2, int x)
{
  return (x-x1)*(y2-y1)/(x2-x1)+y1;
}

uint32 EstimateBitsToCode(const ImageData* img,double q_offset)
{
  uint32 w = img->derivatives.AskWidth();
  uint32 h = img->derivatives.AskHeight();

  uint32 total=0;

  for (uint32 y=0;y<h;y++)
    for (uint32 x=0;x<w;x++)
      {
        int32 d = img->derivatives.Ask_const(y,x);
        int32 dbase = (d-250)/500;

        if (dbase<0)
          dbase=0;
        if (dbase>nDerivValues-2)
          dbase=nDerivValues-2;

        uint32 bits;
        const Macroblock& mb = img->mblks.Ask_const(y,x);

        int q = (int)(mb.m_quantiser_scale+q_offset);

        if (options.CreateMPEG2)
          {
            if (q<  1) q=1;
            if (q>127) q=127;
          }
        else
          {
            if (q< 2) q=2;
            if (q>62) q=62;
          }


        // q auf exp-Skala ummappen

        q = qscale2qcode[1][q];
        assert(q>=1);
        assert(q<=31);

        //cout << "ESTIM " << q << " ";

        if (mb.m_HasMotionForward && mb.m_HasMotionBackward)
          {
            bits = Interpolate(dbase*500+250,dbase*500+750,
                               bittab_D[q][dbase],bittab_D[q][dbase+1],d);
            /*            cout << dbase*500+250 << " " << dbase*500+750 << " "
                 << bittab_D[q][dbase] << " " << bittab_D[q][dbase+1] << " " << d << "   ";
                 cout << bits << " D\n";*/
          }
        else if (mb.m_HasMotionForward)
          {
            bits = Interpolate(dbase*500+250,dbase*500+750,
                               bittab_F[q][dbase],bittab_F[q][dbase+1],d);
            /*            cout << dbase*500+250 << " " << dbase*500+750 << " "
                 << bittab_F[q][dbase] << " " << bittab_D[q][dbase+1] << " " << d << "   ";
                 cout << bits << " F\n";*/
          }
        else if (mb.m_HasMotionBackward)
          {
            bits = Interpolate(dbase*500+250,dbase*500+750,
                               bittab_B[q][dbase],bittab_B[q][dbase+1],d);
            /*            cout << dbase*500+250 << " " << dbase*500+750 << " "
                 << bittab_B[q][dbase] << " " << bittab_D[q][dbase+1] << " " << d << "   ";
                 cout << bits << " B\n";*/
          }
        else
          {
            bits = Interpolate(dbase*500+250,dbase*500+750,
                               bittab_I[q][dbase],bittab_I[q][dbase+1],d);
            /*            cout << dbase*500+250 << " " << dbase*500+750 << " "
                 << bittab_I[q][dbase] << " " << bittab_D[q][dbase+1] << " " << d << "   ";
                 cout << bits << " I\n";*/
          }

        total += bits;
      }

  //  cout << "TOTAL: " << total << endl;

  return total;
}
