/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#ifndef STATISTICS_STATISTICS_HH
#define STATISTICS_STATISTICS_HH

#include "types.hh"
#include "mpeg/constants.hh"
#include "measure/rms.hh"
#include <iostream.h>
#include <iomanip.h>

#define WRITE_STAT 0

class Statistics
{
public:
  void SetCurrentFrameNr(int nr) { d_current_frame=nr; }
  int  AskCurrentFrameNr() const { return d_current_frame; }

  void SetCodingType(uint3 typ) { d_coding_type = typ; }
  void SetPredictionCnt(int intra,int forw,int backw,int bidirect)
    { d_mb_intra=intra; d_mb_forw_pred=forw; d_mb_backw_pred=backw; d_mb_bidirect_pred=bidirect; }
  void SetRMS(double rms) { d_rms=rms; }

  void FinishFrame()
    {
#if WRITE_STAT
      cout << setw(5) << d_current_frame;
      switch (d_coding_type)
        {
        case PICTYPE_I: cout << " I "; break;
        case PICTYPE_P: cout << " P "; break;
        case PICTYPE_B: cout << " B "; break;
        }

      cout << setw(4) << d_mb_intra      << " " << setw(4) << d_mb_forw_pred     << " "
           << setw(4) << d_mb_backw_pred << " " << setw(4) << d_mb_bidirect_pred << " "
           << d_rms << " " << RMS2pSNR(d_rms) << endl;
#endif
    }

private:
  int d_current_frame;

  uint3  d_coding_type;
  short  d_mb_intra;
  short  d_mb_forw_pred;
  short  d_mb_backw_pred;
  short  d_mb_bidirect_pred;
  double d_rms;
};

#endif
